from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.plugin_interp.AbstractInterp import AbstractInterp


class NoopInterp(AbstractInterp):
    """
    Interpreter which does nothing

    It stops processing of the command line on itself (as `next_interp` returns `None` by default)
    """

    def try_iterate(self) -> InterpStep:
        return InterpStep.NextInterp



from argrelay.relay_server.CustomFlaskApp import CustomFlaskApp
from argrelay.relay_server.__main__ import create_app
from argrelay.test_infra.BaseTestClass import BaseTestClass
from argrelay.test_infra.EnvMockBuilder import ServerOnlyEnvMockBuilder


class ServerOnlyTestClass(BaseTestClass):
    """
    Supports FS_66_17_43_42 test_infra / special test mode #7.

    Server-only test without running `argrelay` client at all.

    Note:
    This option is only suitable to run test via `FlaskApp.test_client()`.
    It is not suitable to start server with ports exposed to external clients (e.g. to test GUI).
    To make `FlaskApp` open ports, it requires ugly double-thread-wrapping:
    https://stackoverflow.com/a/56860638/441652
    However, even if it is done to open ports, there is next big problem - there is no known way close the ports.
    There is no known way to control `FlaskApp` instance once it starts (to re-spawn).
    Even shutting down server by throwing exception does not work as the server instance still holds the ports
    open within current test process - new server instance within current test process cannot be re-spawn:
    https://stackoverflow.com/a/42699779/441652
    Exiting current test process prevents all subsequent tests from running - this is not an option either.
    Again, this test class is limited to tests via `FlaskApp.test_client()`.
    To run GUI against the server, use FS_66_17_43_42 test_infra / special test mode #5 instead.
    """

    flask_app: CustomFlaskApp

    @classmethod
    def create_server_in_mocked_env(
        cls,
        env_mock_builder: ServerOnlyEnvMockBuilder,
    ):
        """
        Creates `CustomFlaskApp` started in mocked environment.

        Use case: access FlaskApp.test_client.
        """

        with env_mock_builder.build():
            # This block mocks access to configs while starting Flask.
            # The `env_mock_builder` is ineffective during run of the test cases
            # because this block will already be over.
            # However, the Flask server will still be running for all test cases
            # with the state affected by the mock config data given here.

            cls.flask_app = create_app()

            # Making sure Flask read the mocked server config:
            env_mock_builder.assert_server_config_read()



from marshmallow import Schema, RAISE, fields, post_load

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.schema_config_core_client.ConnectionConfigSchema import connection_config_desc

__comment___ = "__comment__"
connection_config_ = "connection_config"
use_local_requests_ = "use_local_requests"
optimize_completion_request_ = "optimize_completion_request"
show_pending_spinner_ = "show_pending_spinner"
spinless_sleep_sec_ = "spinless_sleep_sec"


# NOTE: Client does not use `Schema` to load config in prod code
#       (only in tests due to heavy import caused by `Schema`).
#       Therefore, validation and applying defaults is not done by this class.
#       Duplicate the same requirements (for validation and defaults) in client prod code.
class ClientConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    # Allow this field in JSON (otherwise schema validation fails):
    __comment__ = fields.String(
        required = False,
    )

    # Serve requests from local data or send to server
    # (used in test only - see `LocalClient` and FS_66_17_43_42 test infra):
    use_local_requests = fields.Boolean(
        required = False,
    )

    # Use one of these (default = True):
    # *   if True: ProposeArgValuesRemoteOptimizedClientCommand
    # *   if False: ProposeArgValuesRemoteClientCommand
    optimize_completion_request = fields.Boolean(
        required = False,
    )

    connection_config = fields.Nested(connection_config_desc.dict_schema)

    # Enables spinner for FS_14_59_14_06: pending requests.
    show_pending_spinner = fields.Boolean(
        required = False,
    )

    spinless_sleep_sec = fields.Number(
        required = False,
        # Noticeable threshold is around 200 ms, but default to spin immediately:
        default = 0.0,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return ClientConfig(
            use_local_requests = input_dict.get(use_local_requests_, False),
            optimize_completion_request = input_dict.get(optimize_completion_request_, True),
            connection_config = input_dict[connection_config_],
            show_pending_spinner = input_dict.get(show_pending_spinner_, False),
            spinless_sleep_sec = input_dict.get(spinless_sleep_sec_, 0.0),
        )


client_config_desc = TypeDesc(
    dict_schema = ClientConfigSchema(),
    ref_name = ClientConfigSchema.__name__,
    dict_example = {
        connection_config_: connection_config_desc.dict_example,
    },
    default_file_path = "argrelay.client.json",
)



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.schema_response.ArgValues import ArgValues


@dataclass
class InterpResult(ArgValues):
    """
    See also :class:`InterpResultSchema`.

    Unlike `InvocationInput` this class contains at most one `data_envelope` per `envelope_container`
    because it executes in the mode which queries values only for latency-sensitive Tab-completion.
    """

    all_tokens: list[str] = field()
    """
    Copy from `ParsedContext.all_tokens` - command line args.
    """

    consumed_tokens: list[int] = field()
    """
    Copy from `InterpContext.consumed_tokens` -
    indexes into `all_tokens` pointing to tokens consumed during interpretation.
    """

    envelope_containers: list[EnvelopeContainer] = field()

    tan_token_ipos: int = field()
    """
    Value from `ParsedContext.tan_token_ipos`.
    """

    tan_token_l_part: str = field()
    """
    Value from `ParsedContext.tan_token_l_part`.
    """

    def get_data_envelopes(self) -> list[dict]:
        data_envelopes = []
        for envelope_container in self.envelope_containers:
            data_envelopes.extend(envelope_container.data_envelopes)
        return data_envelopes



from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.relay_server.QueryEngine import QueryEngine


class HelpHintCache:
    """
    Implements FS_94_30_49_28 help hint.
    """

    def __init__(
        self,
        query_engine: QueryEngine,
    ):
        self.query_engine: QueryEngine = query_engine
        self.help_hint_dict: dict = {}

    def populate_cache(self):

        help_hint_envelopes = self.query_engine.query_data_envelopes(
            ReservedEnvelopeClass.ClassHelp.name,
            {
                f"{ReservedArgType.EnvelopeClass.name}": f"{ReservedEnvelopeClass.ClassHelp.name}",
            },
        )

        for help_hint_envelope in help_hint_envelopes:
            arg_type = help_hint_envelope[ReservedArgType.ArgType.name]
            arg_value = help_hint_envelope[ReservedArgType.ArgValue.name]
            help_hint = help_hint_envelope[ReservedArgType.HelpHint.name]
            if arg_type not in self.help_hint_dict:
                self.help_hint_dict[arg_type] = {}
            self.help_hint_dict[arg_type][arg_value] = help_hint

    def get_value_with_help_hint(
        self,
        arg_type: str,
        arg_value: str,
    ) -> str:
        if arg_type in self.help_hint_dict:
            if arg_value in self.help_hint_dict[arg_type]:
                return f"{arg_value} # {self.help_hint_dict[arg_type][arg_value]}"
        return arg_value



from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.schema_response.InterpResult import InterpResult
from argrelay.schema_response.InterpResultSchema import (
    interp_result_desc,
)
from argrelay.server_spec.CallContext import CallContext


class DescribeLineArgsServerRequestHandler(AbstractServerRequestHandler):

    def __init__(
        self,
        local_server: LocalServer,
    ):
        super().__init__(
            local_server = local_server,
        )

    def handle_request(
        self,
        call_ctx: CallContext,
    ) -> dict:
        assert call_ctx.server_action is ServerAction.DescribeLineArgs

        self.interpret_command(self.local_server, call_ctx)
        ElapsedTime.measure("after_interpret_command")

        input_object = InterpResult(
            arg_values = self.interp_ctx.comp_suggestions,
            all_tokens = self.interp_ctx.parsed_ctx.all_tokens,
            consumed_tokens = self.interp_ctx.consumed_tokens,
            envelope_containers = self.interp_ctx.envelope_containers,
            tan_token_ipos = self.interp_ctx.parsed_ctx.tan_token_ipos,
            tan_token_l_part = self.interp_ctx.parsed_ctx.tan_token_l_part,
        )
        response_dict = interp_result_desc.dict_schema.dump(input_object)

        return response_dict



from copy import deepcopy

from marshmallow import RAISE, fields, pre_dump

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_response.ArgValuesSchema import ArgValuesSchema, arg_values_desc
from argrelay.schema_response.EnvelopeContainerSchema import envelope_container_desc
from argrelay.schema_response.InterpResult import InterpResult

"""
Schema for the result of interpretation taken from :class:`InterpContext`
"""

all_tokens_ = "all_tokens"
consumed_tokens_ = "consumed_tokens"
envelope_containers_ = "envelope_containers"
tan_token_ipos_ = "tan_token_ipos"
tan_token_l_part_ = "tan_token_l_part"


class InterpResultSchema(ArgValuesSchema):
    """
    See also `InterpResult`.
    """

    class Meta:
        unknown = RAISE
        ordered = True

    model_class = InterpResult

    all_tokens = fields.List(
        fields.String(),
        required = True,
    )

    consumed_tokens = fields.List(
        fields.Integer(),
        required = True,
    )

    envelope_containers = fields.List(
        fields.Nested(envelope_container_desc.dict_schema),
        required = True,
    )

    tan_token_ipos = fields.Integer()

    tan_token_l_part = fields.String()

    @pre_dump
    def make_dict(
        self,
        input_object: InterpResult,
        **kwargs,
    ):
        data_dict = super().make_dict(input_object)
        data_dict.update({
            all_tokens_: input_object.all_tokens,
            consumed_tokens_: input_object.consumed_tokens,
            envelope_containers_: input_object.envelope_containers,
            tan_token_ipos_: input_object.tan_token_ipos,
            tan_token_l_part_: input_object.tan_token_l_part,
        })
        return data_dict


_interp_result_example = deepcopy(arg_values_desc.dict_example)
_interp_result_example.update({
    all_tokens_: [
        "some_command",
        "unrecognized_token",
        "goto",
        "host",
        "prod",
    ],
    consumed_tokens_: [
        0,
        2,
        3,
        4,
    ],
    envelope_containers_: [
        envelope_container_desc.dict_example,
    ],
    tan_token_ipos_: 1,
    tan_token_l_part_: "unrecognized_",
})

interp_result_desc = TypeDesc(
    dict_schema = InterpResultSchema(),
    ref_name = InterpResultSchema.__name__,
    dict_example = _interp_result_example,
    default_file_path = "",
)



from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.schema_response.InvocationInput import InvocationInput


class NoopDelegator(AbstractDelegator):

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        invocation_input = InvocationInput(
            arg_values = interp_ctx.comp_suggestions,
            all_tokens = interp_ctx.parsed_ctx.all_tokens,
            consumed_tokens = interp_ctx.consumed_tokens,
            envelope_containers = interp_ctx.envelope_containers,
            tan_token_ipos = interp_ctx.parsed_ctx.tan_token_ipos,
            tan_token_l_part = interp_ctx.parsed_ctx.tan_token_l_part,
            delegator_plugin_entry = local_server.server_config.plugin_instance_entries[
                self.plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        # Do nothing:
        pass



from marshmallow import Schema, fields, RAISE, post_load

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.server_spec.const_int import DEFAULT_IP_ADDRESS, DEFAULT_PORT_NUMBER


class ConnectionConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    server_host_name = fields.String(
        default = DEFAULT_IP_ADDRESS,
    )

    server_port_number = fields.Integer(
        default = DEFAULT_PORT_NUMBER,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return ConnectionConfig(
            server_host_name = input_dict["server_host_name"],
            server_port_number = input_dict["server_port_number"],
        )


connection_config_desc = TypeDesc(
    dict_schema = ConnectionConfigSchema(),
    ref_name = ConnectionConfigSchema.__name__,
    dict_example = {
        "server_host_name": DEFAULT_IP_ADDRESS,
        "server_port_number": DEFAULT_PORT_NUMBER,
    },
    default_file_path = "",
)



from marshmallow import Schema, RAISE, fields, post_load

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_context.InitControl import InitControl

init_types_to_values_ = "init_types_to_values"
"""
Maps types to values for `init_control` (FS_46_96_59_05).
"""


class InitControlSchema(Schema):
    """
    Implements `init_control` (FS_46_96_59_05).
    """

    class Meta:
        unknown = RAISE
        strict = True

    init_types_to_values = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = True,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return InitControl(
            init_types_to_values = input_dict[init_types_to_values_],
        )


init_control_desc = TypeDesc(
    dict_schema = InitControlSchema(),
    ref_name = InitControlSchema.__name__,
    dict_example = {
        init_types_to_values_: {
            "type_a": "TypeA",
            "type_b": "TypeB",
        },
    },
    default_file_path = "",
)



import os
from typing import Union

from argrelay.custom_integ.git_utils import (
    is_git_repo,
    get_full_commit_id,
    get_short_commit_id,
    get_commit_time,
    get_git_repo_root_path,
)
from argrelay.misc_helper_common import get_argrelay_dir
from argrelay.plugin_config.AbstractConfigurator import AbstractConfigurator
from argrelay.plugin_config.DefaultConfiguratorConfigSchema import (
    default_configurator_config_desc,
    commit_id_url_prefix_,
    git_files_by_commit_id_url_prefix_,
    project_title_,
    project_page_url_,
)


class DefaultConfigurator(AbstractConfigurator):

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return default_configurator_config_desc.dict_from_input_dict(plugin_config_dict)

    def provide_project_title(
        self,
    ) -> Union[str, None]:
        if project_title_ in self.plugin_config_dict:
            return self.plugin_config_dict[project_title_]
        else:
            return None

    def provide_project_page_url(
        self,
    ) -> Union[str, None]:
        if project_page_url_ in self.plugin_config_dict:
            return self.plugin_config_dict[project_page_url_]
        else:
            return None

    def provide_project_git_files_by_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        if git_files_by_commit_id_url_prefix_ in self.plugin_config_dict:
            return self.plugin_config_dict[git_files_by_commit_id_url_prefix_]
        else:
            return None

    def provide_project_commit_id_url_prefix(
        self,
    ) -> Union[str, None]:
        if commit_id_url_prefix_ in self.plugin_config_dict:
            return self.plugin_config_dict[commit_id_url_prefix_]
        else:
            return None

    def provide_project_git_commit_id(
        self,
    ) -> Union[str, None]:
        argrelay_dir = get_argrelay_dir()
        if is_git_repo(argrelay_dir):
            return get_full_commit_id(argrelay_dir)
        else:
            return None

    def provide_project_git_commit_display_string(
        self,
    ) -> Union[str, None]:
        argrelay_dir = get_argrelay_dir()
        if is_git_repo(argrelay_dir):
            return get_short_commit_id(argrelay_dir)
        else:
            return None

    def provide_project_git_commit_time(
        self,
    ) -> Union[int, None]:
        argrelay_dir = get_argrelay_dir()
        if is_git_repo(argrelay_dir):
            return get_commit_time(argrelay_dir)
        else:
            return None

    def provide_project_git_repo_relative_argrelay_dir(
        self,
    ) -> Union[str, None]:
        argrelay_dir_abs_path = os.path.realpath(os.path.abspath(get_argrelay_dir()))
        git_repo_root_abs_path = os.path.realpath(os.path.abspath(get_git_repo_root_path(get_argrelay_dir())))
        # Make sure argrelay_dir is not outside the git repo:
        if not argrelay_dir_abs_path.startswith(git_repo_root_abs_path):
            return None
        argrelay_dir_rel_path = os.path.relpath(argrelay_dir_abs_path, git_repo_root_abs_path)
        if argrelay_dir_rel_path == ".":
            # Separator for concatenated path:
            return "/"
        else:
            # Wrap into slashes `/` for concatenation:
            return f"/{argrelay_dir_rel_path}/"

    def provide_project_current_config_path(
        self,
    ) -> Union[str, None]:
        argrelay_dir = get_argrelay_dir()
        conf_path = os.path.join(argrelay_dir, "conf")
        if os.path.islink(conf_path):
            if os.path.isdir(conf_path):
                argrelay_dir_abs_path = os.path.abspath(get_argrelay_dir())
                project_current_config_path = os.readlink(conf_path)
                project_current_config_abs_path = os.path.abspath(os.path.join(
                    argrelay_dir_abs_path,
                    project_current_config_path,
                ))
                if not project_current_config_abs_path.startswith(argrelay_dir_abs_path):
                    return None
                else:
                    return project_current_config_path
            else:
                return None
        elif os.path.isdir(conf_path):
            return conf_path
        else:
            return None



from __future__ import annotations

from marshmallow import Schema, RAISE, fields

from argrelay.custom_integ.FuncConfigSchema import func_config_desc
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.FuncEnvelopeSchema import (
    func_id_some_func_,
)

func_configs_ = "func_configs"


class BaseConfigDelegatorConfigSchema(Schema):
    """
    Part of FS_49_96_50_77 config_only_delegator implementation.
    """

    class Meta:
        unknown = RAISE
        strict = True

    func_configs = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(func_config_desc.dict_schema),
        required = True,
    )


base_config_delegator_config_desc = TypeDesc(
    dict_schema = BaseConfigDelegatorConfigSchema(),
    ref_name = BaseConfigDelegatorConfigSchema.__name__,
    dict_example = {
        func_configs_: {
            func_id_some_func_: func_config_desc.dict_example,
        },
    },
    default_file_path = "",
)



from __future__ import annotations

from copy import deepcopy

from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_interp.FuncTreeInterp import FuncTreeInterp, func_init_control_, func_search_control_
from argrelay.plugin_interp.FuncTreeInterpFactoryConfigSchema import (
    func_selector_tree_,
    func_tree_interp_config_desc,
    jump_tree_,
)
from argrelay.plugin_interp.TreeWalker import TreeWalker
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import init_envelop_collections
from argrelay.schema_config_interp.InitControlSchema import init_types_to_values_
from argrelay.schema_config_interp.SearchControlSchema import (
    keys_to_types_list_,
    populate_search_control,
)
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_enabled_, plugin_dependencies_

tree_path_selector_prefix_ = "tree_path_selector_"
tree_path_selector_key_prefix_ = "s"


class FuncTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_26_43_73_72 func tree.
    Implements FS_91_88_07_23 jump tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        # FS_26_43_73_72 func tree: func id to list of its relative paths populated by `load_func_envelopes`.
        # These func tree paths are relative to the interp node within FS_01_89_09_24 interp tree
        # (as fully qualified func path is composed of interp tree path where this plugin instance is attached to).
        # Each func id can be attached to more than one leaf (hence, there is a list of paths to that func id).
        self.func_ids_to_func_rel_paths: dict[str, list[list[str]]] = {}

        # FS_91_88_07_23 jump tree
        self.plugin_config_dict.setdefault(jump_tree_, {})
        tree_walker: TreeWalker = TreeWalker(
            "jump_tree",
            self.plugin_config_dict[jump_tree_],
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = tree_walker.build_paths_to_paths()
        """
        Implements FS_91_88_07_23 jump tree:
        for given `interp_tree_abs_path` (FS_01_89_09_24) selects next `interp_tree_abs_path`.
        """

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return func_tree_interp_config_desc.dict_from_input_dict(plugin_config_dict)

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        """
        To implement FS_26_43_73_72 func tree, this plugin loads func `data_envelope`-s automatically.

        It loops through func `data_envelope`-s and populates their tree path search props
        according to configured `func_selector_tree` where each function is plugged into.
        """

        mapped_func_ids: list[str] = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        interp_tree_node_config_dict = self.interp_tree_abs_paths_to_node_configs[interp_tree_abs_path]

        func_tree_walker = TreeWalker(
            "func_tree",
            interp_tree_node_config_dict[func_selector_tree_],
        )
        self.func_ids_to_func_rel_paths: dict[str, list[list[str]]] = func_tree_walker.build_str_leaves_paths()

        # Loop through func `data_envelope`-s from all delegators:
        func_envelopes_index: dict[str, dict[tuple[str, ...], dict]] = {}
        for func_id, func_envelope in func_ids_to_func_envelopes.items():
            if func_id not in self.func_ids_to_func_rel_paths:
                # Not used here:
                continue
            if func_id not in func_envelopes_index:
                func_envelopes_index[func_id] = {}
            for func_rel_path in self.func_ids_to_func_rel_paths[func_id]:
                assert tuple(func_rel_path) not in func_envelopes_index[func_id]
                func_envelopes_index[func_id][tuple(func_rel_path)] = deepcopy(func_envelope)
            mapped_func_ids.append(func_id)

        # Ensure every `func_id` mapped into the tree is part of `func_envelope`-s loaded from delegators:
        for func_id in self.func_ids_to_func_rel_paths.keys():
            if func_id not in mapped_func_ids:
                raise RuntimeError(
                    f"plugin_instance_id='{self.plugin_instance_id}': func_id='{func_id}' is not published by any enabled delegator activated so far - please check '{plugin_enabled_}' and '{plugin_dependencies_}'"
                )

        self.populate_init_control(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )

        self.populate_search_control(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )

        (
            interp_tree_abs_path_func_envelopes,
            prop_names,
        ) = self.populate_func_tree_props(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
            func_envelopes_index,
        )

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        class_names = [
            ReservedEnvelopeClass.ClassFunction.name,
        ]
        init_envelop_collections(
            self.server_config,
            class_names,
            lambda collection_name, class_name: (
                [
                    ReservedArgType.EnvelopeClass.name,
                ] + prop_names
            )
        )
        envelope_collection = self.server_config.static_data.envelope_collections[
            class_to_collection_map[ReservedEnvelopeClass.ClassFunction.name]
        ]
        # Write func envelopes into `StaticData` (as if it is a loader plugin):
        envelope_collection.data_envelopes.extend(interp_tree_abs_path_func_envelopes)

        return mapped_func_ids

    # noinspection PyMethodMayBeStatic
    def populate_init_control(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict: dict,
    ):
        interp_tree_node_config_dict[func_init_control_] = {
            init_types_to_values_: {},
        }

        init_types_to_values = interp_tree_node_config_dict[func_init_control_][init_types_to_values_]
        for sel_ipos, path_step in enumerate(interp_tree_abs_path):
            init_types_to_values[f"{func_envelope_path_step_prop_name(sel_ipos)}"] = path_step

    def populate_search_control(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
    ):
        """
        Provide search control based on `func_ids_to_func_rel_paths`.
        """

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        interp_tree_node_config_dict[func_search_control_] = populate_search_control(
            class_to_collection_map,
            ReservedEnvelopeClass.ClassFunction.name,
            [],
        )

        # `func_search_control` should include keys from the interp tree abs path:
        keys_to_types_list = interp_tree_node_config_dict[func_search_control_][keys_to_types_list_]

        # Include interp tree path:
        base_sel_ipos = 0
        for offset_sel_ipos in range(len(interp_tree_abs_path)):
            sel_ipos = base_sel_ipos + offset_sel_ipos
            keys_to_types_list.append({
                f"{path_step_key_arg_name(sel_ipos)}": f"{func_envelope_path_step_prop_name(sel_ipos)}"
            })

        # Include func tree path:
        base_sel_ipos = len(interp_tree_abs_path)
        max_len = max_path_len(self.func_ids_to_func_rel_paths)
        for offset_sel_ipos in range(max_len):
            sel_ipos = base_sel_ipos + offset_sel_ipos
            keys_to_types_list.append({
                f"{path_step_key_arg_name(sel_ipos)}": f"{func_envelope_path_step_prop_name(sel_ipos)}"
            })

    def populate_func_tree_props(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
        func_envelopes_index: dict[str, dict[tuple[str, ...], dict]],
    ) -> tuple[list[dict], list[str]]:
        """
        Populates func tree properties for each func `data_envelope` based on its path in func tree.

        For each func envelope:
        *   Select its place on the func tree (provided by plugin config).
        *   Populate the path as envelope props with `tree_path_selector_prefix_`.
        """
        interp_tree_abs_path_func_envelopes: list[dict] = []
        prop_names: list[str] = []
        for func_id in self.func_ids_to_func_rel_paths:
            func_rel_paths: list[list[str]] = self.func_ids_to_func_rel_paths[func_id]

            for func_rel_path in func_rel_paths:

                func_envelope = func_envelopes_index[func_id][tuple(func_rel_path)]

                # Include interp tree path:
                base_sel_ipos = 0
                for offset_sel_ipos, path_step_id in enumerate(interp_tree_abs_path):
                    sel_ipos = base_sel_ipos + offset_sel_ipos
                    prop_name = func_envelope_path_step_prop_name(sel_ipos)
                    func_envelope[prop_name] = path_step_id
                    prop_names.append(prop_name)

                # Include func tree path:
                base_sel_ipos = len(interp_tree_abs_path)
                for offset_sel_ipos, path_step_id in enumerate(func_rel_path):
                    sel_ipos = base_sel_ipos + offset_sel_ipos
                    prop_name = func_envelope_path_step_prop_name(sel_ipos)
                    func_envelope[prop_name] = path_step_id
                    prop_names.append(prop_name)

                interp_tree_abs_path_func_envelopes.append(func_envelope)

        return (
            interp_tree_abs_path_func_envelopes,
            prop_names,
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        return FuncTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[interp_ctx.interp_tree_abs_path],
            interp_ctx,
            self.func_ids_to_func_rel_paths,
            self.paths_to_jump,
        )


def func_envelope_path_step_prop_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_path_selector_prefix_}{path_step_ipos}"


def path_step_key_arg_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_path_selector_key_prefix_}{path_step_ipos}"


def max_path_len(
    func_ids_to_func_rel_paths: dict[str, list[list[str]]],
):
    """
    Find maximum path length.
    """
    max_len: int = 0
    for func_rel_paths in func_ids_to_func_rel_paths.values():
        for func_rel_path in func_rel_paths:
            max_len = max(max_len, len(func_rel_path))
    return max_len



from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Callable

import jsonpath_ng
from jsonpath_ng import DatumInContext


@dataclass
class JsonTestOutputVerifier:
    """
    Implements TODO_32_99_70_35 verify data via JSONPath queries.
    """

    json_query_assert_list: dict[
        str,
        list[
            Callable[
                [list[DatumInContext]],
                None,
            ]
        ]
    ] = field(default_factory = lambda: {})
    """
    JSON query `str` to list of verifiers as `Callable`-s.

    The `Callable` verifier accepts list query results (to assert anything).
    """

    def add_verifier(
        self,
        json_query: str,
        *assert_func: Callable[[list[DatumInContext]], None],
    ):
        callable_list = self.json_query_assert_list.setdefault(json_query, [])
        for func_item in assert_func:
            callable_list.append(func_item)

        return self

    def verify_all(
        self,
        json_str: str,
    ):
        json_dict = json.loads(json_str)
        for json_query in self.json_query_assert_list:
            jsonpath_expression = jsonpath_ng.parse(json_query)
            json_match = jsonpath_expression.find(json_dict)
            callable_list = self.json_query_assert_list[json_query]
            for callable_item in callable_list:
                callable_item(json_match)



from __future__ import annotations

from marshmallow import Schema, RAISE, fields, post_load

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.ServerPluginControl import ServerPluginControl

first_interp_factory_id_ = "first_interp_factory_id"
reusable_config_data_ = "reusable_config_data"


class ServerPluginControlSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    first_interp_factory_id = fields.String(
        required = True,
    )

    # This `dict` provides a place to store arbitrary data.
    # YAML allows reusing any (substantially complex) data via aliases:
    # https://stackoverflow.com/a/48946813/441652
    # But `marshmallow` does not allow arbitrary data by default
    # (and it is kept that way to let garbage slip through).
    # This `dict` is ignored on load -
    # the data is used by YAML loader before it even reaches the schema validation.
    # See also values merge:
    # https://stackoverflow.com/a/46644785/441652
    reusable_config_data = fields.Dict(
        required = False,
        load_default = {},
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return ServerPluginControl(
            first_interp_factory_id = input_dict[first_interp_factory_id_],
            # `reusable_config_data` is ignored
        )


server_plugin_control_desc = TypeDesc(
    dict_schema = ServerPluginControlSchema(),
    ref_name = ServerPluginControlSchema.__name__,
    dict_example = {
        first_interp_factory_id_: "SomeInterp",
    },
    default_file_path = "",
)



from enum import Enum, auto


class ServiceArgType(Enum):
    """
    Custom arg types used by :class:`ServiceLoader`.

    See also:
    *   `FS_53_81_66_18` (TnC) for arg types.
    *   `TD_63_37_05_36` (demo) for `test_data`.
    """

    # ---

    CodeMaturity = auto()
    GeoRegion = auto()
    FlowStage = auto()

    # ---

    ClusterName = auto()

    # ---

    DataCenter = auto()
    HostName = auto()
    IpAddress = auto()
    GroupLabel = auto()
    ServiceName = auto()

    # ---

    AccessType = auto()

    # ---

    LiveStatus = auto()
    """
    TODO: Currently `LiveStatus` is not used - it can be thought of as manually or dynamically assigned to the resource.
    """

    def __str__(self):
        return self.name



from enum import Enum


class ServerAction(Enum):
    """
    Defines both:
    *   `ServerAction` in REST API URL path for server request (via enum item value)
    *   `ServerAction` in server request payload (via enum item name)
    """

    ProposeArgValues = "/propose_arg_values/"

    DescribeLineArgs = "/describe_line_args/"
    """
    This action is closely related to FS_02_25_41_81 `query_enum_items_func`.
    """

    RelayLineArgs = "/relay_line_args/"



from copy import deepcopy

from marshmallow import RAISE, fields, pre_dump

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import error_delegator_custom_data_desc
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_entry_desc
from argrelay.schema_response.InterpResultSchema import (
    InterpResultSchema,
    interp_result_desc,
)
from argrelay.schema_response.InvocationInput import InvocationInput

delegator_plugin_entry_ = "delegator_plugin_entry"
custom_plugin_data_ = "custom_plugin_data"


class InvocationInputSchema(InterpResultSchema):
    class Meta:
        unknown = RAISE
        ordered = True

    model_class = InvocationInput

    delegator_plugin_entry = fields.Nested(
        plugin_entry_desc.dict_schema,
        required = True,
    )

    custom_plugin_data = fields.Dict(
        required = False,
    )

    @pre_dump
    def make_dict(
        self,
        input_object: InvocationInput,
        **kwargs,
    ):
        data_dict = super().make_dict(input_object)
        data_dict.update({
            delegator_plugin_entry_: input_object.delegator_plugin_entry,
            custom_plugin_data_: input_object.custom_plugin_data,
        })
        return data_dict


_invocation_input_example = deepcopy(interp_result_desc.dict_example)
_invocation_input_example.update({
    delegator_plugin_entry_: plugin_entry_desc.dict_example,
    custom_plugin_data_: error_delegator_custom_data_desc.dict_example,
})

invocation_input_desc = TypeDesc(
    dict_schema = InvocationInputSchema(),
    ref_name = InvocationInputSchema.__name__,
    dict_example = _invocation_input_example,
    default_file_path = "",
)



from marshmallow import Schema, fields, RAISE, post_load, pre_dump

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.PluginEntry import PluginEntry

plugin_enabled_ = "plugin_enabled"
plugin_config_ = "plugin_config"
plugin_module_name_ = "plugin_module_name"
plugin_class_name_ = "plugin_class_name"
plugin_dependencies_ = "plugin_dependencies"


class PluginEntrySchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    plugin_enabled = fields.Boolean(
        required = False,
        load_default = True,
    )

    plugin_module_name = fields.String(
        required = True,
    )

    plugin_class_name = fields.String(
        required = True,
    )

    # List of other plugin instance ids this plugin depends on:
    plugin_dependencies = fields.List(
        fields.String(),
        required = False,
        load_default = [],
    )

    plugin_config = fields.Dict(
        required = False,
        load_default = {},
    )

    @pre_dump
    def make_dict(
        self,
        input_object: PluginEntry,
        **kwargs,
    ):
        return {
            plugin_enabled_: input_object.plugin_enabled,
            plugin_module_name_: input_object.plugin_module_name,
            plugin_class_name_: input_object.plugin_class_name,
            plugin_dependencies_: input_object.plugin_dependencies,
            plugin_config_: input_object.plugin_config,
        }

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return PluginEntry(
            plugin_enabled = input_dict[plugin_enabled_],
            plugin_module_name = input_dict[plugin_module_name_],
            plugin_class_name = input_dict[plugin_class_name_],
            plugin_dependencies = input_dict[plugin_dependencies_],
            plugin_config = input_dict[plugin_config_],
        )


_plugin_entry_example = {
    plugin_enabled_: True,
    plugin_module_name_: "SomePluginModule",
    plugin_class_name_: "SomePluginClass",
    plugin_dependencies_: [],
    plugin_config_: {
    },
}

plugin_entry_desc = TypeDesc(
    dict_schema = PluginEntrySchema(),
    ref_name = PluginEntrySchema.__name__,
    dict_example = _plugin_entry_example,
    default_file_path = "",
)



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.enum_desc.PluginType import PluginType
from argrelay.mongo_data.MongoConfig import MongoConfig
from argrelay.relay_server.GuiBannerConfig import GuiBannerConfig
from argrelay.relay_server.QueryCacheConfig import QueryCacheConfig
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.runtime_data.ServerPluginControl import ServerPluginControl
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_enabled_, plugin_dependencies_


@dataclass
class ServerConfig:
    connection_config: ConnectionConfig = field()
    mongo_config: MongoConfig = field()
    query_cache_config: QueryCacheConfig = field()
    gui_banner_config: GuiBannerConfig = field()
    class_to_collection_map: dict = field()
    server_plugin_control: ServerPluginControl = field()

    plugin_instance_id_activate_list: list[str] = field()
    """
    List of `plugin_instance_id`-s in order of activation generated by
    walking through a DAG described by `plugin_dependencies`.
    Each `plugin_instance_id` is a key into `plugin_instance_entries`.
    """

    plugin_instance_entries: dict[str, PluginEntry] = field()
    """
    Key = `plugin_instance_id`
    """

    # TODO_00_79_72_55: remove in the future:
    static_data: StaticData = field()

    # TODO: Keep this runtime objects in separate (`ServerRuntime`?) class. Ensure/implement ServerConfig dumping on request (for troubleshooting).
    data_loaders: dict[str, "AbstractLoader"] = field(default_factory = lambda: {})
    """
    Entries in `data_loaders` are not directly loaded from config.
    These are plugin instances created during plugin activation.
    """

    # TODO: Keep this runtime objects in separate (`ServerRuntime`?) class. Ensure/implement ServerConfig dumping on request (for troubleshooting).
    interp_factories: dict[str, "AbstractInterpFactory"] = field(default_factory = lambda: {})
    """
    Entries in `interp_factories` are not directly loaded from config.
    These are plugin instances created during plugin activation.
    """

    # TODO: Keep this runtime objects in separate (`ServerRuntime`?) class. Ensure/implement ServerConfig dumping on request (for troubleshooting).
    action_delegators: dict[str, "AbstractDelegator"] = field(default_factory = lambda: {})
    """
    Entries in `action_delegators` are not directly loaded from config.
    These are plugin instances created during plugin activation.
    """

    server_configurators: dict[str, "AbstractConfigurator"] = field(default_factory = lambda: {})


def assert_plugin_instance_id(
    server_config: ServerConfig,
    plugin_instance_id: str,
    plugin_type: PluginType,
):
    error_msg = f"plugin instance `{plugin_instance_id}` must: (A) be `{PluginType.DelegatorPlugin.name}`, (B) be `{plugin_enabled_}`, (C) be activated in the order of DAG via `{plugin_dependencies_}`"
    if plugin_type is PluginType.InterpFactoryPlugin:
        assert plugin_instance_id in server_config.interp_factories, error_msg
    if plugin_type is PluginType.DelegatorPlugin:
        assert plugin_instance_id in server_config.action_delegators, error_msg
    if plugin_type is PluginType.LoaderPlugin:
        assert plugin_instance_id in server_config.data_loaders, error_msg



import importlib
from copy import deepcopy
from typing import Type

from argrelay.enum_desc.PluginType import PluginType
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.runtime_data.ServerConfig import ServerConfig


class AbstractPlugin:

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        self.server_config: ServerConfig = server_config
        self.plugin_instance_id: str = plugin_instance_id
        self.plugin_config_dict: dict = self.load_config(plugin_config_dict)

        self.validate_config()

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        """
        Pre-proces the given config on load (e.g. populate default values).

        The default implementation is to return `plugin_config_dict` as is.

        The typical implementation is to call `TypeDesc.dict_from_input_dict`
        so that defaults are populated according to the given `Schema`.
        """
        return deepcopy(plugin_config_dict)

    def validate_config(
        self,
    ) -> None:
        """
        Validate schema or data for `plugin_config_dict`.

        The typical implementation is to call `TypeDesc.validate_dict(self.plugin_config_dict)`.

        If `load_config` is implemented via `TypeDesc.dict_from_input_dict`,
        validating step is only necessary if it does anything extra beyond what `Schema` can validate.
        """
        pass

    def get_plugin_type(
        self,
    ) -> PluginType:
        """
        Return one of the `PluginType`-s.
        """
        raise NotImplementedError

    def activate_plugin(
        self,
    ) -> None:
        """
        One-time plugin activation callback during server startup.

        This is a chance to initialize (once) any resources plugin requires.
        """
        pass

    def validate_loaded_data(
        self,
        static_data: "StaticData"
    ) -> None:
        """
        Callback to validate data after all :class:`AbstractLoader`-s completed loading data.

        This is a chance for any plugin to verify that the data still matches its expectation.
        """
        pass


def import_plugin_class(
    plugin_entry: PluginEntry,
):
    plugin_module = importlib.import_module(plugin_entry.plugin_module_name)
    plugin_class: Type[AbstractPlugin] = getattr(
        plugin_module,
        plugin_entry.plugin_class_name,
    )
    return plugin_class


def instantiate_plugin(
    server_config: ServerConfig,
    plugin_instance_id: str,
    plugin_entry,
):
    plugin_class = import_plugin_class(plugin_entry)
    plugin_object: AbstractPlugin = plugin_class(
        server_config,
        plugin_instance_id,
        plugin_entry.plugin_config,
    )
    return plugin_object



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.schema_response.InterpResult import InterpResult


@dataclass
class InvocationInput(InterpResult):
    """
    See also :class:`InvocationInputSchema`.

    Unlike `InterpResult` this class provides full list of `data_envelope`-s per `envelope_container`
    because invocation is not latency-sensitive and funcs expect all `data_envelope`-s.
    """

    delegator_plugin_entry: PluginEntry = field()
    """
    The `PluginEntry` taken from config on the server side to let client invoke that plugin.

    It is assumed that server and client code, if not of the same version, at least, compatible.
    """

    custom_plugin_data: dict = field()
    """
    A placeholder dict for exclusive use by plugin.

    Whatever plugin server side needs to tell its plugin peer on the client side.
    """



from copy import deepcopy

from argrelay.custom_integ.ConfigOnlyLoaderConfigSchema import (
    config_only_loader_config_desc,
    data_envelopes_,
    collection_name_to_index_fields_map_,
)
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import init_envelop_collections

_primitive_types = (bool, str, int, float)


def convert_envelope_fields_to_string(
    data_envelope: dict,
):
    """
    Converts primitive types to `str`.

    See also:
    https://stackoverflow.com/a/6392016/441652
    """
    for field_name, field_value in data_envelope.items():
        if isinstance(field_value, _primitive_types):
            data_envelope[field_name] = str(field_value)
    return data_envelope


class ConfigOnlyLoader(AbstractLoader):
    """
    Implements FS_49_96_50_77 config_only_loader.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        self.plugin_config_dict = config_only_loader_config_desc.obj_from_input_dict(self.plugin_config_dict)

    def update_static_data(
        self,
        static_data: StaticData,
    ) -> StaticData:

        # Config for FS_56_43_05_79 different collections:
        class_to_collection_map: dict = self.server_config.class_to_collection_map

        # FS_49_96_50_77 config_only_loader plugin:
        # *   `collection_name_to_index_fields_map`:
        collection_name_to_index_fields_map: dict[str, list[str]] = self.plugin_config_dict[collection_name_to_index_fields_map_]
        # *   list of `data_envelope`-s (actual data):
        data_envelopes = [
            convert_envelope_fields_to_string(data_envelope)
            for data_envelope in deepcopy(self.plugin_config_dict[data_envelopes_])
        ]

        class_names: list[str] = [
            data_envelope.get(ReservedArgType.EnvelopeClass.name, ReservedEnvelopeClass.ClassUnknown.name)
            for data_envelope in data_envelopes
        ]

        init_envelop_collections(
            self.server_config,
            class_names,
            lambda _collection_name, _class_name: collection_name_to_index_fields_map[_collection_name],
        )

        for class_name in class_names:

            class_data_envelopes = static_data.envelope_collections[
                class_to_collection_map[class_name]
            ].data_envelopes

            # Keep removing `data_envelope`-s after inserting them into specific collections based on class:
            filtered_data_envelopes = list([
                data_envelope for data_envelope in data_envelopes
                if data_envelope.get(ReservedArgType.EnvelopeClass.name, None) == class_name
            ])
            class_data_envelopes.extend(filtered_data_envelopes)
            for data_envelope in filtered_data_envelopes:
                data_envelopes.remove(data_envelope)

        if len(data_envelopes) != 0:
            raise RuntimeError(f"ERROR: not all `data_envelope`-s have class assigned: {data_envelopes}")

        return static_data



from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.plugin_interp.FirstArgInterp import FirstArgInterp
from argrelay.plugin_interp.FirstArgInterpFactoryConfigSchema import (
    first_arg_interp_factory_config_desc,
    first_arg_vals_to_next_interp_factory_ids_,
    ignored_func_ids_list_,
)
from argrelay.plugin_interp.InterpTreeInterpFactory import InterpTreeInterpFactory
from argrelay.plugin_interp.InterpTreeInterpFactoryConfigSchema import (
    tree_path_interp_factory_config_desc,
    interp_selector_tree_,
)
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig


class FirstArgInterpFactory(InterpTreeInterpFactory):
    """
    `FirstArgInterp` was re-implemented in terms of `InterpTreeInterp` (FS_01_89_09_24).

    Implements FS_42_76_93_51 first interp.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        first_arg_interp_factory_config_desc.validate_dict(plugin_config_dict)
        converted_config = convert_FirstArgInterpConfig_to_InterpTreeInterpFactoryConfig(plugin_config_dict)
        super().__init__(
            server_config,
            plugin_instance_id,
            converted_config,
        )

    def activate_plugin(
        self,
    ) -> None:
        """
        # NOTE: FS_42_76_93_51 first interp special case:
        TODO: Think of anther way because `FirstArgInterp` is obsolete - this role here only extends its agony.
        `FirstArgInterp` is used as the root and single instance which
        triggers func loading of itself (and all others recursively).

        Takes part in implementation of FS_01_89_09_24 interp tree.
        """

        # Build a map from `func_id` to `func_envelope`:
        func_ids_to_func_envelopes: dict[str, dict] = {}
        for delegator_plugin_id, action_delegator in self.server_config.action_delegators.items():
            func_envelopes = action_delegator.get_supported_func_envelopes()
            for func_envelope in func_envelopes:
                func_id = func_envelope[ReservedArgType.FuncId.name]
                if func_id in func_ids_to_func_envelopes:
                    raise RuntimeError(
                        f"plugin_instance_id='{self.plugin_instance_id}': func_id='{func_id}' is published more than once by delegator_plugin_id='{delegator_plugin_id}'"
                    )
                func_ids_to_func_envelopes[func_id] = func_envelope

        interp_tree_abs_path: tuple[str, ...] = tuple([])
        mapped_func_ids: list[str] = self.load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )

        ignored_func_ids_list = self.plugin_config_dict.get(ignored_func_ids_list_, [])
        for func_id in func_ids_to_func_envelopes.keys():
            if func_id not in mapped_func_ids:
                if func_id not in ignored_func_ids_list:
                    raise RuntimeError(
                        f"plugin_instance_id='{self.plugin_instance_id}': func_id='{func_id}' is neither mapped anywhere nor is in `{ignored_func_ids_list_}`"
                    )
                else:
                    # Func is ignored - skip:
                    continue


    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> FirstArgInterp:
        return FirstArgInterp(
            self.plugin_instance_id,
            # FS_42_76_93_51 first interp special case:
            # We use multiple config clones (with customization) `tree_node_config_dict` for interps at
            # different location on the FS_01_89_09_24 interp tree - these configs get cloned on `load_func_envelopes`.
            # But first interp is always at the root of the interp tree.
            # More over, first interp is the only one who triggers cascading calls to `load_func_envelopes`,
            # but never receives such call itself.
            # Instead of using `interp_tree_node_config_dict`, use `plugin_config_dict` directly:
            self.plugin_config_dict,
            interp_ctx,
        )


def convert_FirstArgInterpConfig_to_InterpTreeInterpFactoryConfig(
    plugin_config_dict: dict,
) -> dict:
    """
    `FirstArgInterp` was re-implemented in terms of `InterpTreeInterp` (FS_01_89_09_24).
    """
    interp_selector_tree = {}
    for command_id, plugin_instance_id in plugin_config_dict[first_arg_vals_to_next_interp_factory_ids_].items():
        interp_selector_tree[command_id] = plugin_instance_id

    output_dict = {
        interp_selector_tree_: interp_selector_tree,
        ignored_func_ids_list_: plugin_config_dict.get(ignored_func_ids_list_, []),
    }

    tree_path_interp_factory_config_desc.validate_dict(output_dict)

    return output_dict



from __future__ import annotations

import os
from dataclasses import dataclass, field

from argrelay.enum_desc.CompScope import CompScope
from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper_common import eprint
from argrelay.server_spec.CallContext import CallContext

UNKNOWN_COMP_KEY: str = str(0)
"""
Constant used when client is called via `CallConv.CliArgsConv`
and there is no `COMP_KEY` env var to obtain the value
(unlike during `CallConv.EnvVarsConv`).
"""

COMP_LINE_env_var: str = "COMP_LINE"
COMP_POINT_env_var: str = "COMP_POINT"
COMP_TYPE_env_var: str = "COMP_TYPE"
COMP_KEY_env_var: str = "COMP_KEY"


@dataclass(frozen = True)
class ShellContext:
    """
    Immutable input data from a shell
    """

    command_line: str = field()
    cursor_cpos: int = field()
    comp_type: CompType = field()
    comp_key: str = field()
    is_debug_enabled: bool = field()

    def __post_init__(self):
        assert 0 <= self.cursor_cpos <= len(self.command_line)

    @classmethod
    def from_env(
        cls,
        argv: list[str],
    ) -> ShellContext:
        """
        See Bash docs on these env var:
        https://www.gnu.org/software/bash/manual/html_node/Bash-Variables.html
        """
        if COMP_LINE_env_var in os.environ:
            # See `CallConv.EnvVarsConv`:
            command_line = os.environ[COMP_LINE_env_var]
            cursor_cpos = int(os.environ[COMP_POINT_env_var])
            # If `COMP_LINE_env_var` exists, the call is definitely NOT for `CompType.InvokeAction`:
            comp_type = CompType(int(os.environ[COMP_TYPE_env_var]))
            assert comp_type != CompType.InvokeAction
            comp_key = os.environ[COMP_KEY_env_var]
        else:
            # See `CallConv.CliArgsConv`:
            argv = [os.path.basename(argv[0])] + argv[1:]
            command_line = " ".join(argv)
            cursor_cpos = len(command_line)
            # If `COMP_LINE_env_var` is missing, the call is definitely for `CompType.InvokeAction`
            comp_type = CompType.InvokeAction
            comp_key = UNKNOWN_COMP_KEY

        is_debug_enabled = (
            "ARGRELAY_DEBUG" in os.environ
            and
            "p" in os.environ["ARGRELAY_DEBUG"]
        )

        return cls(
            command_line = command_line,
            cursor_cpos = cursor_cpos,
            comp_type = comp_type,
            comp_key = comp_key,
            is_debug_enabled = is_debug_enabled,
        )

    def print_debug(
        self,
        end_str: str = "\n",
    ) -> None:
        if not self.is_debug_enabled:
            return
        eprint(TermColor.debug_output.value, end = "")
        eprint(f"\"{self.command_line}\"", end = " ")
        eprint(f"cursor_cpos: {self.cursor_cpos}", end = " ")
        eprint(f"comp_type: {self.comp_type}", end = " ")
        eprint(TermColor.reset_style.value, end = end_str)

    def create_call_context(
        self,
    ) -> CallContext:
        server_action: ServerAction = self.select_server_action()
        return CallContext(
            server_action = server_action,
            command_line = self.command_line,
            cursor_cpos = self.cursor_cpos,
            comp_scope = CompScope.from_comp_type(self.comp_type),
            is_debug_enabled = self.is_debug_enabled,
        )

    def select_server_action(
        self,
    ) -> ServerAction:
        if self.comp_type is CompType.DescribeArgs:
            return ServerAction.DescribeLineArgs
        if self.comp_type is CompType.InvokeAction:
            return ServerAction.RelayLineArgs

        return ServerAction.ProposeArgValues



import time
import uuid
from copy import deepcopy

from pymongo import MongoClient

from argrelay.enum_desc.PluginType import PluginType
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.misc_helper_common import eprint
from argrelay.mongo_data import MongoClientWrapper
from argrelay.mongo_data.MongoServerWrapper import MongoServerWrapper
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.relay_server.HelpHintCache import HelpHintCache
from argrelay.relay_server.QueryEngine import QueryEngine
from argrelay.runtime_context.AbstractPlugin import instantiate_plugin, AbstractPlugin
from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.StaticDataSchema import static_data_desc


class LocalServer:
    """
    This is plain server functionality without API-wrapper to expose over the network (hence, local).

    The API-wrapper exposing `LocalServer` over the network is `CustomFlaskApp`.
    """

    def __init__(
        self,
        server_config: ServerConfig,
    ):
        self.server_instance_id = uuid.uuid4()
        self.server_config: ServerConfig = server_config
        self.mongo_server: MongoServerWrapper = MongoServerWrapper()
        self.mongo_client: MongoClient = MongoClientWrapper.get_mongo_client(self.server_config.mongo_config)
        self.query_engine: QueryEngine = QueryEngine(
            self.server_config.query_cache_config,
            self.get_mongo_database(),
            self.server_config.mongo_config.distinct_values_query,
        )
        self.help_hint_cache: HelpHintCache = HelpHintCache(
            self.query_engine,
        )
        # seconds since epoch:
        self.server_start_time: int = int(time.time())

    def start_local_server(self):
        self._activate_plugins()
        self._start_mongo_server()
        self._load_mongo_data()
        self._create_mongo_index()
        self._populate_help_hint_cache()
        self._log_connection_url()

    def stop_local_server(self):
        self._stop_mongo_server()

    def get_mongo_database(self):
        return self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

    def get_query_engine(self):
        return self.query_engine

    def _activate_plugins(self):
        """
        Calls each plugin to update :class:`StaticData`.
        """

        for plugin_instance_id in self.server_config.plugin_instance_id_activate_list:
            plugin_entry: PluginEntry = self.server_config.plugin_instance_entries[plugin_instance_id]

            if not plugin_entry.plugin_enabled:
                continue

            plugin_instance: AbstractPlugin = instantiate_plugin(
                self.server_config,
                plugin_instance_id,
                plugin_entry,
            )
            plugin_type = plugin_instance.get_plugin_type()

            if plugin_type is PluginType.LoaderPlugin:
                plugin_instance: AbstractLoader
                plugin_instance.activate_plugin()
                # Store instance of `AbstractLoader` under specified id for future use:
                self.server_config.data_loaders[plugin_instance_id] = plugin_instance
                # Use loader to update data:
                self.server_config.static_data = plugin_instance.update_static_data(self.server_config.static_data)
                continue

            if plugin_type is PluginType.InterpFactoryPlugin:
                plugin_instance: AbstractInterpFactory
                plugin_instance.activate_plugin()
                # Store instance of `AbstractInterpFactory` under specified id for future use:
                self.server_config.interp_factories[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.DelegatorPlugin:
                plugin_instance: AbstractDelegator
                plugin_instance.activate_plugin()
                # Store instance of `AbstractDelegator` under specified id for future use:
                self.server_config.action_delegators[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.ConfiguratorPlugin:
                plugin_instance: AbstractDelegator
                plugin_instance.activate_plugin()
                # Store instance of `AbstractConfigurator` under specified id for future use:
                self.server_config.server_configurators[plugin_instance_id] = plugin_instance
                continue

        eprint("validating data...")
        self._validate_static_data()

    def _validate_static_data(self):
        self._validate_static_data_schema()
        self._validata_static_data_by_plugins()

    def _validate_static_data_schema(self):
        # Note that this is slow for large data sets:
        static_data_dict = static_data_desc.dict_schema.dump(self.server_config.static_data)
        static_data_desc.validate_dict(static_data_dict)

    def _validata_static_data_by_plugins(self):
        all_plugins: [AbstractLoader] = []
        all_plugins.extend(self.server_config.data_loaders.values())
        all_plugins.extend(self.server_config.interp_factories.values())
        all_plugins.extend(self.server_config.action_delegators.values())

        for plugin_instance in all_plugins:
            plugin_instance.validate_loaded_data(self.server_config.static_data)

    def _start_mongo_server(self):
        self.mongo_server.start_mongo_server(self.server_config.mongo_config)

    def _stop_mongo_server(self):
        self.mongo_server.stop_mongo_server()

    def _load_mongo_data(self):
        mongo_db = self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]
        MongoClientWrapper.store_envelopes(
            mongo_db,
            # TODO_00_79_72_55: Remove `static_data` from `server_config`:
            self.server_config.static_data,
        )

    def _create_mongo_index(self):
        mongo_db = self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

        for collection_name in self.server_config.static_data.envelope_collections:
            envelope_collection: EnvelopeCollection = self.server_config.static_data.envelope_collections[
                collection_name
            ]
            # Include `envelope_class` field into index by default:
            index_fields: list[str] = deepcopy(envelope_collection.index_fields)
            index_fields.append(ReservedArgType.EnvelopeClass.name)
            MongoClientWrapper.create_index(
                mongo_db,
                collection_name,
                envelope_collection.index_fields,
            )

    def _populate_help_hint_cache(self):
        self.help_hint_cache.populate_cache()

    def _log_connection_url(self):
        host_name = self.server_config.connection_config.server_host_name
        port_number = self.server_config.connection_config.server_port_number
        eprint(f"http://{host_name}:{port_number}")



"""
Mock environment (env vars, command line args, file access, database, etc.) for `argrelay` client or server
"""

from __future__ import annotations

import contextlib
import dataclasses
import io
import json
import os
import re
import sys
import tempfile
from contextlib import ExitStack
from dataclasses import dataclass, field
from io import StringIO
from typing import Type, Union, Callable
from unittest import mock

import yaml

from argrelay.client_spec.ShellContext import (
    UNKNOWN_COMP_KEY,
    ShellContext,
    COMP_LINE_env_var,
    COMP_POINT_env_var,
    COMP_TYPE_env_var,
    COMP_KEY_env_var,
)
from argrelay.custom_integ.GitRepoLoader import GitRepoLoader
from argrelay.custom_integ.ServiceLoader import ServiceLoader
from argrelay.custom_integ.ServiceLoaderConfigSchema import test_data_ids_to_load_
from argrelay.enum_desc.CallConv import CallConv
from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.DistinctValuesQuery import DistinctValuesQuery
from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.misc_helper_common import get_argrelay_dir
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.runtime_context.ParsedContext import ParsedContext
from argrelay.schema_config_core_client.ClientConfigSchema import (
    use_local_requests_,
    client_config_desc,
    optimize_completion_request_,
    show_pending_spinner_,
)
from argrelay.schema_config_core_server.MongoConfigSchema import (
    use_mongomock_,
    distinct_values_query_,
)
from argrelay.schema_config_core_server.QueryCacheConfigSchema import enable_query_cache_
from argrelay.schema_config_core_server.ServerConfigSchema import (
    mongo_config_,
    server_config_desc,
    plugin_instance_entries_,
    query_cache_config_,
)
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_config_, plugin_enabled_
from argrelay.schema_response.InvocationInput import InvocationInput
from argrelay.server_spec.CallContext import CallContext
from argrelay.test_infra.LocalClientCommandFactory import LocalClientCommandFactory
from argrelay.test_infra.OpenFileMock import OpenFileMock
from argrelay.test_infra.PopenMock import PopenMock


@dataclass
class EnvMockBuilder:
    """
    All-in-one mock support which sets up the mocks and cleans them up as Python's Context Manager.

    For example:

    *   Mock env vars or `sys.argv` used by Bash to communicate input to argrelay client - see usage of:

        *   `_mock_client_input_in_completion_mode`
        *   `_mock_client_input_in_invocation_mode_with_args`
        *   `_mock_client_input_in_invocation_mode_with_line`

    *   Mock server and client config files - see usage of:

        *   `set_server_config_dict`
        *   `set_client_config_dict`

    *   Capture `stdout` and `stderr` - see usage of:

        *   `set_capture_stdout`
        *   `set_capture_stderr`

    *   Whether client uses `LocalClient`/`LocalServer` or `RemoteClient` with `CustomFlaskApp` - see usage of:

        *   `set_client_config_with_local_server`

    *   Mock MongoDB client - see usage of: `use_mongomock`

    *   Simple selection of test data - see usage of: `set_test_data_ids_to_load`

    *   TODO_42_81_01_90: Verify `ArgValues` on `ServerAction.DescribeLineArgs` (without verifying printed output) by intercepting call to `ProposeArgValuesClientResponseHandler.render_values`.
    *   TODO_42_81_01_90: Verify `InterpResult` on `ServerAction.ProposeArgValues` (without verifying printed output) by intercepting call to `DescribeLineArgsClientResponseHandler.render_result`.
    *   Verifying `InvocationInput` on `ServerAction.RelayLineArgs` - see usage of:

        *   `invoke_action_func_full_name`

    *   ...

    """

    ####################################################################################################################
    # Client input

    command_line: Union[str, None] = field(default = None)
    command_args: Union[list[str], None] = field(default = None)
    cursor_cpos: Union[int, None] = field(default = None)
    comp_type: Union[CompType, None] = field(default = None)
    # The `comp_key` value is not used at the moment (but it is provided by shell in case of `CallConv.EnvVarsConv`) -
    # using any default value for now to avoid setting it everywhere
    # (when it starts to matter, change default to `None` to force review and specifying it in all tests):
    comp_key: str = field(default = UNKNOWN_COMP_KEY)

    # TODO: Keep default to False (to do the minimum by default) - one has to select pre-set builder class to get defaults instead:
    _mock_client_input: bool = field(default = True)

    ####################################################################################################################
    # Client config

    is_client_config_with_local_server: bool = field(default = True)
    is_client_config_to_optimize_completion_request: bool = field(default = None)
    show_pending_spinner: Union[bool, None] = field(default = False)

    ####################################################################################################################
    # Server config

    test_data_ids_to_load: list[str] = field(default_factory = lambda: [
        "TD_70_69_38_46",  # no data
    ])

    enable_demo_git_loader: bool = field(default = False)
    use_mongomock: Union[bool, None] = field(default = None)

    enable_query_cache: Union[bool, None] = field(default = None)
    """
    See `QueryCacheConfig.enable_query_cache`.
    """

    distinct_values_query: Union[DistinctValuesQuery, None] = field(default = None)
    """
    See `DistinctValuesQuery`.
    """

    ####################################################################################################################
    # Output capturing

    actual_stdout: StringIO = field(default = None)
    capture_stdout: bool = field(default = False)

    actual_stderr: StringIO = field(default = None)
    capture_stderr: bool = field(default = False)

    ####################################################################################################################
    # Config file mocking

    client_config_dict: Union[dict, None] = field(default_factory = lambda: None)
    server_config_dict: Union[dict, None] = field(default_factory = lambda: None)

    mock_client_config_file_read: bool = field(default = False)
    mock_server_config_file_read: bool = field(default = False)

    # Implement FS_62_25_92_06 generated config for `out`-processes to read it (see FS_66_17_43_42 test infra):
    generate_client_config_file: bool = field(default = False)
    generate_server_config_file: bool = field(default = False)
    temp_test_config_dir: Union[tempfile.TemporaryDirectory, None] = None

    ####################################################################################################################
    # General mocking

    file_mock: OpenFileMock = field(default_factory = lambda: OpenFileMock({}))

    assert_on_close: bool = field(default = True)

    ####################################################################################################################
    # Intercept specific payload

    invoke_action_func_full_name: str = field(default = None)
    """
    Set by giving `delegator_class` to `set_capture_delegator_invocation_input`.
    """
    invocation_input: InvocationInput = field(default = None)
    """
    Captured `InvocationInput` by using `capture_invocation_input` func on `ServerAction.RelayLineArgs`
    instead of calling client-side delegator.
    """

    ####################################################################################################################
    # Server control

    reset_local_server: bool = field(default = True)
    """
    If true, (after build() context is over) next invocation via `LocalClient` will trigger `LocalServer` restart (re-creation and re-load).
    Default is true because it is confusing to hold `test_data_ids_to_load` while not re-loading server by default.
    """

    was_server_started_on_build: bool = field(default = False)
    """
    Avoids triggering verification of file access mock usage for server config
    when `LocalClient` reuses already running server.
    See FS_66_17_43_42 test infra for server-`out` test modes.
    """

    ####################################################################################################################
    # Client input

    def set_command_line(self, command_line: str):
        """
        Used as input in case of `CallConv.CliArgsConv` because `COMP_LINE` is env var what Bash sets
        """
        self.command_line = command_line
        return self

    def set_command_args(self, command_args: list[str]):
        """
        Used as input in case of `CallConv.EnvVarsConv` because list[str] is what `sys.argv` provides
        """
        self.command_args = command_args
        return self

    def set_cursor_cpos(self, cursor_cpos: int):
        self.cursor_cpos = cursor_cpos
        return self

    def set_comp_type(self, comp_type: CompType):
        self.comp_type = comp_type
        return self

    def set_comp_key(self, comp_key: str):
        self.comp_key = comp_key
        return self

    def set_mock_client_input(self, given_val: bool):
        self._mock_client_input = given_val
        return self

    ####################################################################################################################
    # Client config

    def set_client_config_with_local_server(self, given_val: bool):
        self.is_client_config_with_local_server = given_val
        return self

    def set_client_config_to_optimize_completion_request(self, given_val: Union[bool, None]):
        self.is_client_config_to_optimize_completion_request = given_val
        return self

    def set_show_pending_spinner(
        self,
        given_val: Union[bool, None],
    ):
        self.show_pending_spinner = given_val
        return self

    ####################################################################################################################
    # Server config

    def set_test_data_ids_to_load(self, test_data_ids_to_load: list[str]):
        self.test_data_ids_to_load = test_data_ids_to_load
        return self

    def set_enable_demo_git_loader(self, given_val: bool):
        self.enable_demo_git_loader = given_val
        return self

    def set_use_mongomock(
        self,
        given_val: Union[bool, None],
    ):
        self.use_mongomock = given_val
        return self

    def set_enable_query_cache(
        self,
        given_val: Union[bool, None],
    ):
        self.enable_query_cache = given_val
        return self

    def set_distinct_values_query(
        self,
        distinct_values_query: Union[distinct_values_query, None]
    ):
        self.distinct_values_query = distinct_values_query
        return self

    ####################################################################################################################
    # Output capturing

    def set_capture_stdout(self, given_val: bool):
        self.capture_stdout = given_val
        return self

    def set_capture_stderr(self, given_val: bool):
        self.capture_stderr = given_val
        return self

    ####################################################################################################################
    # Config file mocking

    def set_client_config_dict(
        self,
        client_config: Union[dict, None] = None,
    ):
        if client_config is None:
            client_config = client_config_desc.dict_from_default_file()
        self.client_config_dict = client_config
        return self

    def set_server_config_dict(
        self,
        server_config: Union[dict, None] = None,
    ):
        if server_config is None:
            server_config = server_config_desc.dict_from_default_file()
        self.server_config_dict = server_config
        return self

    def get_client_config_json(self):
        return json.dump(self.client_config_dict)

    def get_server_config_yaml(self):
        return yaml.dump(self.server_config_dict)

    def set_mock_client_config_file_read(self, given_val: bool):
        self.mock_client_config_file_read = given_val
        return self

    def set_mock_server_config_file_read(self, given_val: bool):
        self.mock_server_config_file_read = given_val
        return self

    def set_generate_client_config_file(self, given_val: bool):
        self.generate_client_config_file = given_val
        return self

    def set_generate_server_config_file(self, given_val: bool):
        self.generate_server_config_file = given_val
        return self

    ####################################################################################################################
    # General mocking

    @contextlib.contextmanager
    def mock_file_open(self):
        with mock.patch("builtins.open", self.file_mock.open) as file_mock:
            yield file_mock

    ####################################################################################################################
    # Intercept specific payload

    def set_capture_delegator_invocation_input(self, delegator_class: Type[AbstractDelegator]):
        """
        This func causes `AbstractDelegator.invoke_action` to be mocked to capture `InvocationInput`
        inside `EnvMockBuilder.invocation_input` allowing tests to assert
        the data received from server on `ServerAction.RelayLineArgs`.
        """

        self.invoke_action_func_full_name = (
            f"{delegator_class.__module__}"
            "."
            f"{delegator_class.__name__}"
            "."
            "invoke_action"
        )
        return self

    ####################################################################################################################
    # Server control

    def set_reset_local_server(self, given_val: bool):
        self.reset_local_server = given_val
        return self

    ####################################################################################################################

    @contextlib.contextmanager
    def build(self):
        """
        Mock resources via "Combining Multiple Context Managers":
        https://rednafi.github.io/digressions/python/2020/03/26/python-contextmanager.html#combining-multiple-context-managers

        Warning:
            The approach below works in tests for setting up and tearing down mocks (judging by behavior),
            but there is no guarantee it is robust under all scenarios.
            For real resource management (when various exceptions raised), it has to be thoroughly tested.
        """

        ################################################################################################################
        # Initial validation

        self.was_server_started_on_build = LocalClientCommandFactory.local_server is not None

        # Ensure there are no false expectations if conflicting setup is done:
        assert self.command_line is None or self.command_args is None, "both cannot be set, at most one"

        if self.command_line is not None:
            assert self.cursor_cpos is not None, "setting command line (in CompletionMode) requires setting cursor cpos"

        if self.command_args is not None:
            assert self.cursor_cpos is None, "if args are set (in InvocationMode), cursor pos should not be set"

        self.assert_config_substitution_mutually_exclusive()

        ################################################################################################################
        # Client config

        if self.client_config_dict is not None:
            self.assert_client_config_substitute_enabled()

        if self.is_client_config_with_local_server:
            # So far, local client is only used for testing (which implies using mocked client config file access).
            # If fails here, for consistency, either enable client config file mocking or disable local client.
            self.assert_client_config_substitute_enabled()

        if self.is_client_config_to_optimize_completion_request is not None:
            self.assert_client_config_substitute_enabled()

        if self.show_pending_spinner is not None:
            self.assert_client_config_substitute_enabled()

        ################################################################################################################
        # Server config

        if self.server_config_dict is not None:
            self.assert_server_config_substitute_enabled()

        if self.enable_demo_git_loader:
            self.assert_server_config_substitute_enabled()

        if self.enable_query_cache is not None:
            self.assert_server_config_substitute_enabled()

        if self.distinct_values_query is not None:
            self.assert_server_config_substitute_enabled()

        if self.use_mongomock is not None:
            self.assert_server_config_substitute_enabled()

        ################################################################################################################

        if self.mock_client_config_file_read or self.generate_client_config_file:

            assert self.client_config_dict is not None

            self.client_config_dict[use_local_requests_] = self.is_client_config_with_local_server

            if self.is_client_config_to_optimize_completion_request is not None:
                self.client_config_dict[
                    optimize_completion_request_
                ] = self.is_client_config_to_optimize_completion_request

            if self.show_pending_spinner is not None:
                self.client_config_dict[
                    show_pending_spinner_
                ] = self.show_pending_spinner

            # set mocked file content:
            if self.mock_client_config_file_read:
                self.file_mock.path_to_data[client_config_desc.get_adjusted_file_path()] = json.dumps(
                    self.client_config_dict
                )

        if self.mock_server_config_file_read or self.generate_server_config_file:
            """
            Change server config data, then mock file access to return that data for tests.
            """
            assert self.server_config_dict is not None

            # TODO: Do not hardcode plugin id (instance of `GitRepoLoader`):
            plugin_entry = self.server_config_dict[plugin_instance_entries_][f"{GitRepoLoader.__name__}.default"]
            plugin_entry[plugin_enabled_] = self.enable_demo_git_loader

            # TODO: Do not hardcode plugin id (instance of `ServiceLoader`):
            plugin_entry = self.server_config_dict[plugin_instance_entries_][f"{ServiceLoader.__name__}.default"]
            plugin_entry[plugin_config_][test_data_ids_to_load_] = self.test_data_ids_to_load

            if self.enable_query_cache is not None:
                self.server_config_dict[query_cache_config_][enable_query_cache_] = self.enable_query_cache

            if self.distinct_values_query is not None:
                self.server_config_dict[mongo_config_][distinct_values_query_] = self.distinct_values_query.name

            if self.use_mongomock is not None:
                self.server_config_dict[mongo_config_][use_mongomock_] = self.use_mongomock

            # set mocked file content:
            if self.mock_server_config_file_read:
                self.file_mock.path_to_data[server_config_desc.get_adjusted_file_path()] = yaml.dump(
                    self.server_config_dict
                )

        with ExitStack() as exit_stack:

            # noinspection PyListCreation
            yield_list = []

            # Always mock file access - whether file data or mocked data is given depends on the mock config:
            yield_list.append(exit_stack.enter_context(self.mock_file_open()))

            if self._mock_client_input:

                assert self.comp_type is not None, "if mocking client input was requested, `comp_type` cannot be `None`"

                if CallConv.from_comp_type(self.comp_type) is CallConv.EnvVarsConv:
                    if self.command_line is not None and self.cursor_cpos is not None:
                        # TODO: make explicit function "mock_client_input_in_env_vars" with all three args required.
                        yield_list.append(exit_stack.enter_context(
                            _mock_client_input_in_env_vars(
                                self.command_line,
                                self.cursor_cpos,
                                self.comp_type,
                            )
                        ))
                    else:
                        raise RuntimeError
                elif CallConv.from_comp_type(self.comp_type) is CallConv.CliArgsConv:
                    # TODO: make explicit function "mock_client_input_in_cli_args" with just command_args.
                    if self.command_args is not None:
                        # TODO: do not branch here, branch on mock setup (in client tests) to make it explicit/conscious that InvocationMode is not about command_line, but command_args.
                        yield_list.append(exit_stack.enter_context(
                            _mock_client_input_in_invocation_mode_with_args(
                                self.command_args,
                            )
                        ))
                    elif self.command_line is not None:
                        # TODO: do not branch here, branch on mock setup (in client tests) to make it explicit/conscious that InvocationMode is not about command_line, but command_args.
                        yield_list.append(exit_stack.enter_context(
                            _mock_client_input_in_invocation_mode_with_line(
                                self.command_line,
                            )
                        ))
                    else:
                        raise RuntimeError
                else:
                    raise RuntimeError

            if self.capture_stdout:
                self.actual_stdout = io.StringIO()
                yield_list.append(exit_stack.enter_context(_capture_stdout(self.actual_stdout)))

            if self.capture_stderr:
                self.actual_stderr = io.StringIO()
                yield_list.append(exit_stack.enter_context(_capture_stderr(self.actual_stderr)))

            if self.assert_on_close:
                yield_list.append(exit_stack.enter_context(self.assert_all_cm()))

            if self.invoke_action_func_full_name:
                yield_list.append(exit_stack.enter_context(
                    _mock_delegator_plugin(self.invoke_action_func_full_name)
                ))

            if self.reset_local_server:
                yield_list.append(exit_stack.enter_context(
                    do_reset_local_server()
                ))

            if (
                self.generate_client_config_file
                or
                self.generate_server_config_file
            ):
                yield_list.append(exit_stack.enter_context(
                    self.generate_configs()
                ))

            yield yield_list

    def assert_config_substitution_mutually_exclusive(self):
        assert (
            (
                self.mock_client_config_file_read == self.generate_client_config_file == False
                or
                self.mock_client_config_file_read != self.generate_client_config_file
            )
            and
            (
                self.mock_server_config_file_read == self.generate_server_config_file == False
                or
                self.mock_server_config_file_read != self.generate_server_config_file
            )
        ), "FS_62_25_92_06: generated config: mocked config and generated config are mutually exclusive"

    def assert_client_config_substitute_enabled(self):
        assert (
            self.mock_client_config_file_read
            or
            self.generate_client_config_file
        ), "FS_62_25_92_06: generated config: when enabled: either mocked config or generated config"

    def assert_server_config_substitute_enabled(self):
        assert (
            self.mock_server_config_file_read
            or
            self.generate_server_config_file
        ), "FS_62_25_92_06: generated config: when enabled: either mocked config or generated config"

    @contextlib.contextmanager
    def assert_all_cm(self):
        try:
            yield
        finally:
            if self.mock_client_config_file_read:
                self.assert_client_config_read()
            if self.mock_server_config_file_read and not self.was_server_started_on_build:
                self.assert_server_config_read()

    def assert_client_config_read(self):
        self.assert_file_read(client_config_desc.get_adjusted_file_path())

    def assert_server_config_read(self):
        self.assert_file_read(server_config_desc.get_adjusted_file_path())

    def assert_file_read(self, file_path: str):
        """
        Ensures that mocked file was actually accessed.

        If fails here, it means either/or:
        *   test setup was over-mocked (e.g. see `mock_server_config_file_read`, `mock_client_config_file_read`)
        *   test did not hit functionality that is supposed to access the file
        """
        self.file_mock.path_to_mock[file_path].assert_called_with(file_path)

    @contextlib.contextmanager
    def generate_configs(
        self,
    ):
        base_tmp_dir = f"{get_argrelay_dir()}/tmp"
        assert os.path.isdir(base_tmp_dir)

        test_configs_dir = f"{base_tmp_dir}/test_configs"
        if not os.path.exists(test_configs_dir):
            os.makedirs(test_configs_dir)

        self.temp_test_config_dir = tempfile.TemporaryDirectory(dir = f"{test_configs_dir}/")
        try:
            with self.temp_test_config_dir:
                os.environ["ARGRELAY_CONF_BASE_DIR"] = self.temp_test_config_dir.name
                self.inner_generate_configs()
                yield
        finally:
            os.environ.pop("ARGRELAY_CONF_BASE_DIR")

    def inner_generate_configs(self):
        client_config_path = os.path.join(
            self.temp_test_config_dir.name,
            client_config_desc.default_file_path,
        )
        server_config_path = os.path.join(
            self.temp_test_config_dir.name,
            server_config_desc.default_file_path,
        )
        if self.generate_client_config_file:
            self.generate_client_config(str(client_config_path))
        if self.generate_server_config_file:
            self.generate_server_config(str(server_config_path))

    def generate_client_config(
        self,
        file_path: str,
    ):
        with open(file_path, "w") as json_file:
            json.dump(
                self.client_config_dict,
                json_file,
            )

    def generate_server_config(
        self,
        file_path: str,
    ):
        with open(file_path, "w") as yaml_file:
            yaml.dump(
                self.server_config_dict,
                yaml_file,
            )


########################################################################################################################
# Pre-configured mock builders

@dataclass
class EmptyEnvMockBuilder(EnvMockBuilder):
    """
    Use case:
    Used to set up extra mocks before or after when nesting `EnvMockBuilder` one into another (or completely alone).
    Without calling any extra setters after creation to prime some mocks, this `EnvMockBuilder`  is noop.
    """

    def __init__(
        self,
    ):
        super().__init__()
        self.set_mock_client_input(False)
        self.set_reset_local_server(False)
        # Disable all mocks which set tripwires if not used:
        self.set_mock_client_config_file_read(False)
        self.set_mock_server_config_file_read(False)
        self.set_client_config_with_local_server(False)
        # Client config is not substituted but otherwise default `show_pending_spinner = False` requires it - use None:
        self.set_show_pending_spinner(None)


@dataclass
class LocalClientEnvMockBuilder(EnvMockBuilder):
    """
    Supports FS_66_17_43_42 test_infra / special test mode #1.

    Use case:
    Used in tests where both server and client code is verified but without code for data marshalling via HTTP.
    Runs client and server code in the same test process via `LocalClient` (see for more details).
    """

    def __init__(
        self,
    ):
        super().__init__()

        # TODO: enable validation that client code is actually invoked (e.g. invocation of make_request in main):

        # Ensure that client and server read their config files by test process:
        self.set_mock_client_config_file_read(True)
        self.set_mock_server_config_file_read(True)
        self.set_generate_client_config_file(False)
        self.set_generate_server_config_file(False)

        # Load default configs as file access mocks input:
        self.set_client_config_dict()
        self.set_server_config_dict()

        # For local client (with local server) tests,
        # ensure client uses `LocalClient` without marshalling data via HTTP:
        self.set_client_config_with_local_server(True)
        # Also, ensure non-optimized completion is used
        # (optimized directly uses network socket which will not give access to `LocalServer`):
        self.set_client_config_to_optimize_completion_request(False)

        # For local client (with local server) tests,
        # client code will need to access data passed by the shell - mock it:
        self.set_mock_client_input(True)


@dataclass
class ServerOnlyEnvMockBuilder(EnvMockBuilder):
    """
    Supports FS_66_17_43_42 test_infra / special test mode #7.

    Use case:
    Used in tests where client code is not invoked (or external to the test process like GUI).
    Current test process runs only server code - use this class to mock server environment (e.g. config).
    """

    def __init__(
        self,
    ):
        super().__init__()

        # TODO: enable validation that client code is not invoked (e.g. invocation of make_request):

        # For server-only test, client config file read should not happen by test process
        # (if this mock is enabled, but not used, it fails):
        self.set_mock_client_config_file_read(False)
        self.set_generate_client_config_file(False)

        # For server-only test, use mocked server config file access (to control its content in test):
        self.set_mock_server_config_file_read(True)
        self.set_generate_server_config_file(False)

        # Load default server config as file access mock input:
        self.set_server_config_dict()

        # For server-only test, client code is not used, but if it is, try to fail via REST API:
        self.set_client_config_with_local_server(False)

        # For server-only test, client input mocking is not required:
        self.set_mock_client_input(False)

        # Client config is not substituted but otherwise default `show_pending_spinner = False` requires it - use None:
        self.set_show_pending_spinner(None)


@dataclass
class LiveServerEnvMockBuilder(EnvMockBuilder):
    """
    Supports FS_66_17_43_42 test_infra / special test mode #2.

    Use case:
    Used in tests where client talks to some live server.
    Server is started somehow outside the mock, current test process runs only client code.
    """

    def __init__(
        self,
    ):
        super().__init__()

        # TODO: enable validation that server code is not invoked (e.g. no invocation of create_call_ctx):

        # For live server test, do not substitute client config by default
        # (because default client config contains correct connection details for default live server):
        self.set_mock_client_config_file_read(False)
        self.set_generate_client_config_file(False)
        # Client config is not substituted but otherwise default `show_pending_spinner = False` requires it - use None:
        self.set_show_pending_spinner(None)

        # For live server test, server config file substitution should not happen by test process code
        # (because it happens within server process - un-verify-able from the test process):
        self.set_mock_server_config_file_read(False)
        self.set_generate_server_config_file(False)

        # For live server test, running local server contradicts with the live server - disable:
        self.set_client_config_with_local_server(False)
        # Optimized client prevents accessing internal server state via `LocalClient` and `LocalServer`,
        # but it is impossible with live server anyway.
        # Therefore, default is unspecified here = None
        # (whatever is set in the client config or dictated by its defaults):
        self.set_client_config_to_optimize_completion_request(None)


########################################################################################################################
# Client input mocking

@contextlib.contextmanager
def _mock_client_input_in_env_vars(command_line: str, cursor_cpos: int, comp_type: CompType):
    with mock.patch.dict(os.environ, {
        COMP_LINE_env_var: command_line,
        COMP_POINT_env_var: str(cursor_cpos),
        COMP_TYPE_env_var: str(comp_type.value),
        COMP_KEY_env_var: UNKNOWN_COMP_KEY,
    }) as env_mock:
        yield env_mock


def _mock_client_input_in_invocation_mode_with_line(command_line: str):
    command_args = re.compile(SpecialChar.TokenDelimiter.value).split(command_line)
    return _mock_client_input_in_invocation_mode_with_args(command_args)


@contextlib.contextmanager
def _mock_client_input_in_invocation_mode_with_args(command_args: list[str]):
    with mock.patch.object(sys, "argv", command_args) as argv_mock:
        yield argv_mock


########################################################################################################################
# Output capturing

@contextlib.contextmanager
def _capture_stdout(stdout_f):
    with contextlib.redirect_stdout(stdout_f) as stdout_mock:
        yield stdout_mock


@contextlib.contextmanager
def _capture_stderr(stderr_f):
    with contextlib.redirect_stderr(stderr_f) as stderr_mock:
        yield stderr_mock


########################################################################################################################
# Intercept specific payload

@contextlib.contextmanager
def _mock_delegator_plugin(path_to_invoke_action):
    with mock.patch(path_to_invoke_action, capture_invocation_input) as invoke_action_mock:
        yield invoke_action_mock


def capture_invocation_input(invocation_input: InvocationInput):
    """
    This body substitutes (mocks) `invoke_action` func in `DelegatorPlugin`-s.

    Instead of executing func logic, it only captures its `InvocationInput` for verifications in tests.
    """
    EnvMockBuilder.invocation_input = dataclasses.replace(invocation_input)


########################################################################################################################
# Stand-alone util mock funcs

@contextlib.contextmanager
def mock_subprocess_popen(expected_args_to_output):
    with mock.patch("subprocess.Popen", PopenMock(expected_args_to_output)) as popen_mock:
        yield popen_mock


@contextlib.contextmanager
def wrap_instance_method_on_instance(
    any_instance,
    callable_on_instance: Callable,
):
    """
    Wraps `callable_on_instance` by a mock which still calls original method.
    Use case: test whether original method is called.
    See also: `wrap_instance_method_on_class` - similar purpose but requires completely different mocking call.
    """
    with mock.patch.object(
        any_instance,
        callable_on_instance.__name__,
        wraps = callable_on_instance,
    ) as method_wrap_mock:
        yield method_wrap_mock


@contextlib.contextmanager
def wrap_instance_method_on_class(
    any_class,
    callable_on_instance: Callable,
):
    """
    Wraps `callable_on_instance` by a mock which still calls original method.
    Use case: test whether original method is called.
    See also: `wrap_instance_method_on_instance` - similar purpose but requires completely different mocking call.
    """
    with mock.patch(
        f"{any_class.__module__}.{any_class.__name__}.{callable_on_instance.__name__}",
        side_effect = callable_on_instance,
        autospec = True,
    ) as method_wrap_mock:
        yield method_wrap_mock


@contextlib.contextmanager
def do_reset_local_server():
    try:
        yield
    finally:
        LocalClientCommandFactory.local_server = None


########################################################################################################################
# Test objects

def default_test_parsed_context(
    command_line: str,
    cursor_cpos: int,
) -> ParsedContext:
    return ParsedContext.from_instance(
        default_test_input_context(
            command_line,
            cursor_cpos,
        ),
    )


def default_test_input_context(
    command_line: str,
    cursor_cpos: int,
) -> CallContext:
    return ShellContext(
        command_line = command_line,
        cursor_cpos = cursor_cpos,
        comp_type = CompType.PrefixShown,
        comp_key = UNKNOWN_COMP_KEY,
        is_debug_enabled = False,
    ).create_call_context()



from __future__ import annotations

from typing import Union, Type

from argrelay.client_command_local.AbstractLocalClientCommand import AbstractLocalClientCommand
from argrelay.enum_desc.CompType import CompType
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.relay_client import __main__
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.schema_response.ArgValuesSchema import arg_values_
from argrelay.server_spec.CallContext import CallContext
from argrelay.test_infra import parse_line_and_cpos
from argrelay.test_infra.EnvMockBuilder import EnvMockBuilder, LocalClientEnvMockBuilder
from argrelay.test_infra.InOutTestClass import InOutTestClass
from argrelay.test_infra.LocalClientCommandFactory import LocalClientCommandFactory


class LocalTestClass(InOutTestClass):
    """
    Supports FS_66_17_43_42 test_infra / special test mode #1.

    Tests which rely on `LocalClient` to inspect internal-to-server `InterpContext`.
    """

    @classmethod
    def setUpClass(cls):
        InOutTestClass.setUpClass()
        # Force restart of the server for `LocalClient` before tests:
        LocalClientCommandFactory.local_server = None

    @classmethod
    def tearDownClass(cls):
        InOutTestClass.tearDownClass()
        # Force restart of the server for `LocalClient` after tests:
        LocalClientCommandFactory.local_server = None

    def verify_output_with_new_server_via_local_client(
        self,
        test_data: str,
        test_line: str,
        comp_type: CompType,
        expected_suggestions: Union[list[str], None],
        container_ipos_to_expected_assignments: Union[dict[int, dict[str, AssignedValue]], None],
        delegator_class: Union[Type[AbstractDelegator], None],
        envelope_ipos_to_field_values: Union[dict[int, dict[str, str]], None],
    ):
        self.verify_output_with_via_local_client(
            test_data,
            test_line,
            comp_type,
            expected_suggestions,
            container_ipos_to_expected_assignments,
            None,
            delegator_class,
            envelope_ipos_to_field_values,
            LocalClientEnvMockBuilder(),
        )

    def verify_output_with_via_local_client(
        self,
        test_data: str,
        test_line: str,
        comp_type: CompType,
        expected_suggestions: Union[list[str], None],
        container_ipos_to_expected_assignments: Union[dict[int, dict[str, AssignedValue]], None],
        container_ipos_to_options_hidden_by_default_value: Union[dict[int, dict[str, list[str]]], None],
        delegator_class: Union[Type[AbstractDelegator], None],
        envelope_ipos_to_field_values: Union[dict[int, dict[str, str]], None],
        init_env_mock_builder: EnvMockBuilder,
    ):
        (command_line, cursor_cpos) = parse_line_and_cpos(test_line)

        if envelope_ipos_to_field_values is not None:
            self.assertIsNotNone(delegator_class)

        if delegator_class is not None:
            init_env_mock_builder.set_capture_delegator_invocation_input(delegator_class)

        env_mock_builder = (
            init_env_mock_builder
            .set_command_line(command_line)
            .set_cursor_cpos(cursor_cpos)
            .set_comp_type(comp_type)
            .set_test_data_ids_to_load([
                test_data,
            ])
        )
        with env_mock_builder.build():
            command_obj = __main__.main()
            assert isinstance(command_obj, AbstractLocalClientCommand)

            # This server-side internal data `InterpContext` is only available in tests
            # when `LocalClient` and `LocalServer` are used.
            interp_ctx: InterpContext = command_obj.interp_ctx
            call_ctx: CallContext = command_obj.call_ctx

            # TODO_32_99_70_35: (JSONPath?) Currently, this verifier ensures what things exists.
            #       Add a way to ensure what things do not exist.
            #       For example,
            #       * no `data_envelope` number N.
            #       * `data_envelope` does not have field A.
            #       * `data_envelope` does not have value 'x' for field A.

            actual_suggestions = None
            if expected_suggestions is not None:
                actual_suggestions = command_obj.response_dict[arg_values_]

            self.verify_response_data(
                call_ctx,
                actual_suggestions,
                container_ipos_to_expected_assignments,
                container_ipos_to_options_hidden_by_default_value,
                delegator_class,
                envelope_ipos_to_field_values,
                expected_suggestions,
                interp_ctx.envelope_containers,
            )



from __future__ import annotations

from collections.abc import Callable
from copy import deepcopy
from typing import Union

default_tree_leaf_ = ""


class TreeWalker:
    """
    Helper with logic to traverse tree - used to implement:
    *   FS_01_89_09_24 interp tree
    *   FS_26_43_73_72 func tree
    *   FS_91_88_07_23 jump tree
    *   FS_33_76_82_84 global tree

    The tree is specified as `dict` (normally, in config).
    """

    def __init__(
        self,
        tree_name: str,
        tree_dict: dict,
    ):
        self.tree_name = tree_name
        self.tree_dict = tree_dict
        self.leaf_type = None

    def build_str_leaves_paths(
        self,
    ) -> dict[str, list[list[str]]]:
        """
        For each named leaf (becomes key) provide its paths in the tree.

        Each named leaf can have several paths in the tree (leaves with the same name).

        Use case:
        *   We know `func_id`, we need to know which FS_26_43_73_72 func tree paths it is registered under.
        """
        tree_paths: dict[str, list[list[str]]] = {}
        self.leaf_type = str
        self._build_leaves_sub_paths(
            [],
            self.tree_dict,
            tree_paths,
            self._handle_str_leaf,
        )
        return tree_paths

    # TODO: Unused (and, apparently, it wasn't used in the past).
    def build_tuple_leaves_paths(
        self,
    ) -> dict[tuple[str, ...], list[list[str]]]:
        """
        For each leaf value (`list` in tree leaf becomes `tuple` key in the map) provide its paths in the tree.

        Each leaf (its `tuple` converted from `list`) can have several paths in the tree (leaves with the same value).

        In tree, leaf is a `list`.
        In map, leaf becomes `tuple` (as a key for `dict`).

        Use case:
        *   We know tree path (tuple) defined in the nodes, we need to know all tree paths where same path is defined.
            (not a real use case as it is unused TODO: clean up)
        """
        tree_paths: dict[tuple[str, ...], list[list[str]]] = {}
        self.leaf_type = list
        self._build_leaves_sub_paths(
            [],
            self.tree_dict,
            tree_paths,
            self._handle_list_leaf,
        )
        return tree_paths

    def build_paths_to_paths(
        self
    ) -> dict[tuple[str, ...], tuple[str, ...]]:
        """
        Map each path in the tree into path specified in the leaf (`list` in tree leaf becomes `tuple` key in the map).

        In tree, leaf is a `list`.
        In map, leaf becomes `tuple` (as a key for `dict`).

        Use case:
        *   We know curr tree path (tuple), we need to know next tree path where to jump (FS_91_88_07_23 jump tree).
        """

        paths_to_paths: dict[tuple[str, ...], tuple[str, ...]] = {}
        self.leaf_type = list
        self._build_paths_to_paths(
            [],
            self.tree_dict,
            paths_to_paths,
        )
        return paths_to_paths

    def _build_leaves_sub_paths(
        self,
        curr_path: list[str],
        sub_tree: Union[str, list, dict],
        output_leaves_paths: dict[Union[str, tuple[str, ...]], list[list[str]]],
        handle_leaf: Callable[
            [
                list[str],
                Union[str, list, dict],
                dict[Union[str, tuple[str, ...]],
                list[list[str]]],
            ],
            None,
        ],
    ):
        if isinstance(sub_tree, dict):
            for node_id in sub_tree:
                tree_node = sub_tree[node_id]
                if node_id == default_tree_leaf_:
                    self._build_leaves_sub_paths(
                        curr_path,
                        tree_node,
                        output_leaves_paths,
                        handle_leaf,
                    )
                else:
                    next_leaf_path = deepcopy(curr_path)
                    next_leaf_path.append(node_id)
                    self._build_leaves_sub_paths(
                        next_leaf_path,
                        tree_node,
                        output_leaves_paths,
                        handle_leaf,
                    )
        elif isinstance(sub_tree, self.leaf_type):
            handle_leaf(
                curr_path,
                sub_tree,
                output_leaves_paths,
            )
        else:
            self._raise_on_wrong_type(curr_path, sub_tree)

    def _build_paths_to_paths(
        self,
        curr_path: list[str],
        sub_tree: Union[list, dict],
        output_paths_to_paths: dict[tuple[str, ...], tuple[str, ...]],
    ):
        if isinstance(sub_tree, dict):
            for node_id in sub_tree:
                tree_node = sub_tree[node_id]
                if node_id == default_tree_leaf_:
                    self._build_paths_to_paths(
                        curr_path,
                        tree_node,
                        output_paths_to_paths,
                    )
                else:
                    next_leaf_path = deepcopy(curr_path)
                    next_leaf_path.append(node_id)
                    self._build_paths_to_paths(
                        next_leaf_path,
                        tree_node,
                        output_paths_to_paths,
                    )
        elif isinstance(sub_tree, self.leaf_type):
            self._verify_list_str(curr_path, sub_tree)
            output_paths_to_paths[tuple(curr_path)] = tuple(sub_tree)
        else:
            self._raise_on_wrong_type(curr_path, sub_tree)

    def _handle_str_leaf(
        self,
        curr_path: list[str],
        leaf_node: str,
        output_leaves_paths: dict[str, list[list[str]]],
    ) -> None:
        if leaf_node == default_tree_leaf_:
            if len(curr_path) < 1:
                raise ValueError(
                    f"{self.tree_name}: tree path `{curr_path}` must contain at least one step to use leaf `{default_tree_leaf_}`"
                )
            leaf_id = curr_path[-1]
            leaf_path = curr_path[:-1]
        else:
            leaf_id = leaf_node
            leaf_path = deepcopy(curr_path)

        if leaf_id in output_leaves_paths:
            output_leaves_paths[leaf_id].append(leaf_path)
        else:
            output_leaves_paths[leaf_id] = [leaf_path]

    def _handle_list_leaf(
        self,
        curr_path: list[str],
        leaf_node: list,
        output_leaves_paths: dict[tuple[str, ...], list[list[str]]],
    ) -> None:
        """
        Convert `list` leaf node to `tuple` to be used as key in `output_leaves_paths` `dict`.
        """
        self._verify_list_str(curr_path, leaf_node)

        leaf_id = tuple(leaf_node)
        leaf_path = deepcopy(curr_path)

        if leaf_id in output_leaves_paths:
            output_leaves_paths[leaf_id].append(leaf_path)
        else:
            output_leaves_paths[leaf_id] = [leaf_path]

    def _verify_list_str(
        self,
        curr_path: list[str],
        list_value: list,
    ):
        if not all(isinstance(leaf_list_item, str) for leaf_list_item in list_value):
            raise ValueError(
                f"{self.tree_name}: tree path `{curr_path}` leaf list must contain only `str`"
            )

    def _raise_on_wrong_type(
        self,
        curr_path: list[str],
        typed_value,
    ):
        raise ValueError(
            f"{self.tree_name}: tree path `{curr_path}` leaf type `{typed_value.__class__.__name__}` is neither `{self.leaf_type.__name__}` nor `dict`"
        )



"""
Internal constants
"""

# See `server_title` and `server_version` instead.
UNUSED_TITLE = "argrelay"
UNUSED_VERSION = "0.0.0"

# This will make `apispec` generate schema compatible with `flasgger` (placing schemas under `#/definitions/`):
DEFAULT_OPEN_API_VERSION = "2.0"

# Built-in GUI:
ARGRELAY_GUI_PATH = "/argrelay_gui/"

# Run API docs GUI at the root:
API_DOCS_PATH = "/argrelay_api/"

# main:
DEFAULT_IP_ADDRESS = "localhost"
DEFAULT_PORT_NUMBER = 8787

API_SPEC_PATH = "/argrelay_server_api_spec.json"

# noinspection HttpUrlsUsage
BASE_URL_FORMAT = "http://{server_host_name}:{server_port_number}"



from __future__ import annotations

from typing import Union

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator, get_func_id_from_invocation_input
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.plugin_interp.TreeWalker import TreeWalker
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.InterpContext import InterpContext, function_container_ipos_
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_response.InvocationInput import InvocationInput

# TODO: Add config schema for both `InterceptDelegator` and `HelpDelegator`:
tree_abs_path_to_interp_id_ = "tree_abs_path_to_interp_id"


class InterceptDelegator(AbstractDelegator):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        func_tree_walker = TreeWalker(
            tree_abs_path_to_interp_id_,
            self.plugin_config_dict[tree_abs_path_to_interp_id_],
        )
        # Temporary (reversed) map which contains path per id (instead of id per path):
        temporary_id_to_paths: dict[str, list[list[str]]] = func_tree_walker.build_str_leaves_paths()

        # Reverse temporary path per id map into id per path map:
        self.tree_path_to_next_interp_plugin_instance_id: dict[tuple[str, ...], str] = {}
        for interp_plugin_id, tree_abs_paths in temporary_id_to_paths.items():
            for tree_abs_path in tree_abs_paths:
                self.tree_path_to_next_interp_plugin_instance_id[tuple(tree_abs_path)] = interp_plugin_id

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        func_envelopes = [{
            instance_data_: {
                func_id_: SpecialFunc.intercept_invocation_func.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                ],
            },
            ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedArgType.HelpHint.name: (
                f"Intercept and print `{InvocationInput.__name__}` "
                "for specified function and its args"
            ),
            ReservedArgType.FuncId.name: SpecialFunc.intercept_invocation_func.name,
        }]
        return func_envelopes

    def run_init_control(
        self,
        envelope_containers: list[EnvelopeContainer],
        curr_container_ipos: int,
    ):
        """
        `InterceptDelegator` searches its own func envelope.
        """
        super().run_init_control(
            envelope_containers,
            curr_container_ipos,
        )
        curr_container = envelope_containers[curr_container_ipos]
        curr_container.assigned_types_to_values[
            ReservedArgType.FuncId.name
        ] = AssignedValue(
            SpecialFunc.intercept_invocation_func.name,
            ArgSource.InitValue,
        )

    def run_search_control(
        self,
        function_data_envelope: dict,
    ) -> list[SearchControl]:
        # Nothing to search (only if next interpreter may need more, but this one is done):
        return []

    def run_interp_control(
        self,
        curr_interp: AbstractInterp,
    ) -> Union[None, str]:
        # TODO_10_72_28_05: support special funcs for all commands:
        #                   Delegator must select next `interp_factory_id` based on `interp_tree_abs_path` via `tree_abs_path_to_interp_id` (not based on single plugin id specified)
        #                   because delegator can be accessible through multiple tree paths.
        #                   Delegator should map specific tree path using the tree path they are accessed through to specific `interp_plugin_instance_id`.
        #                   The next interp selection is a double jump:
        #                   *   the first jump here (in delegator) is based on selected func via its delegator to interp,
        #                   *   the second jump there (in jump tree interp) from jump interp via jump tree.
        # TODO: This must be special interpreter which is configured only to search functions (without their args).
        #       NEXT TODO: Why not support function args for special interp (e.g. `intercept` or `help` with format params)?
        #                  The interp_control is only run when func and all its args are selected and there should be next interp to continue.
        if curr_interp.interp_ctx.interp_tree_abs_path in self.tree_path_to_next_interp_plugin_instance_id:
            return self.tree_path_to_next_interp_plugin_instance_id[curr_interp.interp_ctx.interp_tree_abs_path]
        else:
            return None

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        # TODO: Fail (send to ErrorDelegator) if next function is not specified -
        #       showing the payload in this case is misleading.
        function_envelope = interp_ctx.envelope_containers[function_container_ipos_]
        delegator_plugin_instance_id = (
            function_envelope
            .data_envelopes[0]
            [instance_data_][delegator_plugin_instance_id_]
        )
        invocation_input = InvocationInput(
            arg_values = interp_ctx.comp_suggestions,
            all_tokens = interp_ctx.parsed_ctx.all_tokens,
            consumed_tokens = interp_ctx.consumed_tokens,
            envelope_containers = interp_ctx.envelope_containers,
            tan_token_ipos = interp_ctx.parsed_ctx.tan_token_ipos,
            tan_token_l_part = interp_ctx.parsed_ctx.tan_token_l_part,
            delegator_plugin_entry = local_server.server_config.plugin_instance_entries[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        if get_func_id_from_invocation_input(invocation_input) == SpecialFunc.intercept_invocation_func.name:
            # TODO: Print without first function `data_envelope` belonging to `intercept` function:
            print(invocation_input)



from marshmallow import Schema, RAISE, fields, post_load

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.plugin_config.DefaultConfiguratorConfig import DefaultConfiguratorConfig

"""
Schema for :class:`DefaultConfiguratorConfig`
"""

project_title_ = "project_title"

project_page_url_ = "project_page_url"

git_files_by_commit_id_url_prefix_ = "git_files_by_commit_id_url_prefix"

commit_id_url_prefix_ = "commit_id_url_prefix"


class DefaultConfiguratorConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        ordered = True

    project_title = fields.String(
        required = False,
    )

    project_page_url = fields.String(
        required = False,
    )

    git_files_by_commit_id_url_prefix = fields.String(
        required = False,
    )

    commit_id_url_prefix = fields.String(
        required = False,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return DefaultConfiguratorConfig(
            project_title = input_dict.get(project_title_, None),
            project_page_url = input_dict.get(project_page_url_, None),
            git_files_by_commit_id_url_prefix = input_dict.get(git_files_by_commit_id_url_prefix_, None),
            commit_id_url_prefix = input_dict.get(commit_id_url_prefix_, None),
        )


default_configurator_config_desc = TypeDesc(
    dict_schema = DefaultConfiguratorConfigSchema(),
    ref_name = DefaultConfiguratorConfigSchema.__name__,
    dict_example = {
        project_title_: "relay_demo",
        project_page_url_: "https://argrelay.org",
        git_files_by_commit_id_url_prefix_: "https://github.com/argrelay/argrelay/tree/",
        commit_id_url_prefix_: "https://github.com/argrelay/argrelay/commit/",
    },
    default_file_path = "",
)



# Keep minimal import on start:
import sys

from argrelay.client_spec.ShellContext import ShellContext
from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.misc_helper_common import get_config_path
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig

ElapsedTime.measure("after_program_entry")


def main():
    ElapsedTime.measure("after_initial_imports")

    file_path = get_config_path("argrelay.client.json")
    client_config = load_client_config(file_path)
    ElapsedTime.measure("after_loading_client_config")

    shell_ctx = ShellContext.from_env(sys.argv)
    shell_ctx.print_debug()
    call_ctx = shell_ctx.create_call_context()

    if (
        # FS_14_59_14_06 pending requests: at the moment the process is split only to provide spinner:
        client_config.show_pending_spinner
        and
        shell_ctx.comp_type is not CompType.InvokeAction
    ):
        from argrelay.relay_client.proc_split import split_process
        (
            is_parent,
            child_pid,
            child_stdout,
        ) = split_process()

        if is_parent:
            from argrelay.relay_client.proc_parent import (
                print_child_stdout_chunks,
                read_next_child_stdout_chunk,
                spin_wait_for_child,
            )
            spin_wait_for_child(
                child_pid,
                child_stdout,
                shell_ctx,
                client_config.spinless_sleep_sec,
            )

            while read_next_child_stdout_chunk(child_stdout):
                pass

            print_child_stdout_chunks()
            return

    if client_config.use_local_requests:
        # This branch with `use_local_requests` is used only for testing
        # (to inspect internal server data via `LocalClient` and `LocalServer`):
        from argrelay.test_infra.LocalClient import LocalClient

        command_obj = make_request_via_abstract_client(
            LocalClient(client_config),
            call_ctx,
        )
    else:
        if (
            call_ctx.server_action is ServerAction.ProposeArgValues
            and
            client_config.optimize_completion_request
        ):
            from argrelay.client_command_remote.ProposeArgValuesRemoteOptimizedClientCommand import (
                ProposeArgValuesRemoteOptimizedClientCommand
            )
            command_obj = ProposeArgValuesRemoteOptimizedClientCommand(
                call_ctx = call_ctx,
                connection_config = client_config.connection_config,
            )
            make_request(
                command_obj,
            )
        else:
            from argrelay.relay_client.RemoteClient import RemoteClient
            command_obj = make_request_via_abstract_client(
                RemoteClient(client_config),
                call_ctx,
            )

    ElapsedTime.measure("on_exit")
    if shell_ctx.is_debug_enabled:
        ElapsedTime.print_all()

    return command_obj


def load_client_config(file_path):
    import json
    with open(file_path) as config_file:
        client_config_dict = json.load(config_file)
    client_config = client_config_dict_to_object(client_config_dict)
    return client_config


def client_config_dict_to_object(client_config_dict):
    """
    Optimized dict -> object conversion avoiding import of `*Schema` for performance:
    """
    client_config = ClientConfig(
        use_local_requests = client_config_dict.get("use_local_requests", False),
        optimize_completion_request = client_config_dict.get("optimize_completion_request", True),
        connection_config = ConnectionConfig(
            server_host_name = client_config_dict["connection_config"]["server_host_name"],
            server_port_number = client_config_dict["connection_config"]["server_port_number"],
        ),
        show_pending_spinner = client_config_dict.get("show_pending_spinner", True),
        spinless_sleep_sec = client_config_dict.get("spinless_sleep_sec", 0.0),
    )
    return client_config


def make_request_via_abstract_client(
    abstract_client: "AbstractClient",
    call_ctx,
) -> "AbstractClientCommand":
    ElapsedTime.measure("before_client_invocation")
    return abstract_client.make_request(call_ctx)


def make_request(
    command_obj,
):
    command_obj.execute_command()
    ElapsedTime.measure("after_execute_command")
    return command_obj


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection


@dataclass(frozen = True)
class StaticData:
    """
    All data which normally does not change between requests.

    See also `StaticDataSchema`.
    """

    envelope_collections: dict[str, EnvelopeCollection] = field()
    """
    MondoDB collection name to its `EnvelopeCollection`.

    Collection name is normally (but not necessarily) matching one of the `ReservedArgType.EnvelopeClass`.
    """



from __future__ import annotations

from marshmallow import Schema, RAISE, fields

from argrelay.custom_integ.GitRepoArgType import GitRepoArgType
from argrelay.misc_helper_common.TypeDesc import TypeDesc

repo_base_path_ = "repo_base_path"
repo_rel_path_ = "repo_rel_path"
is_repo_enabled_ = "is_repo_enabled"
envelope_properties_ = "envelope_properties"


class GitRepoEntryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    repo_base_path = fields.String()

    repo_rel_path = fields.String()

    is_repo_enabled = fields.Boolean(
        load_default = True,
        required = False,
    )

    # These key-values are copied into `data_envelope`:
    envelope_properties = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = True,
    )


git_repo_entry_config_desc = TypeDesc(
    dict_schema = GitRepoEntryConfigSchema(),
    ref_name = GitRepoEntryConfigSchema.__name__,
    dict_example = {
        repo_rel_path_: "argrelay.git",
        is_repo_enabled_: True,
        envelope_properties_: {
            GitRepoArgType.GitRepoAlias.name: "ar",
        }
    },
    default_file_path = "",
)



from enum import Enum, auto


class SpecialFunc(Enum):
    """
    See also: FS_80_45_89_81 / integrated functions
    """

    intercept_invocation_func = auto()
    """
    Implements FS_88_66_66_73 intercept func.
    """

    help_hint_func = auto()
    """
    Implements FS_71_87_33_52 help hint.
    """

    echo_args_func = auto()
    """
    Implements FS_43_50_57_71 `echo_args` func.
    """

    query_enum_items_func = auto()
    """
    Implements FS_02_25_41_81 `query_enum_items_func`.
    """



from enum import Enum, auto


class InterpStep(Enum):
    StopAll = auto()
    """
    No way to continue interpretation - no enough info.
    """

    NextEnvelope = auto()
    """
    Current search yields single `data_envelope`, move to search the next envelope.
    """

    NextInterp = auto()
    """
    All `data_envelope`-s for current interpreter have been found, switch to the next interpreter.
    """



class AbstractClientResponseHandler:

    def __init__(
        self,
    ):
        pass

    def handle_response(
        self,
        response_dict: dict,
    ):
        raise NotImplementedError



from __future__ import annotations

from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.FuncEnvelopeSchema import (
    func_envelope_desc,
)

func_envelope_ = "func_envelope"
fill_control_list_ = "fill_control_list"

class FuncConfigSchema(Schema):
    """
    Part of FS_49_96_50_77 config_only_plugins implementation.
    """

    class Meta:
        unknown = RAISE
        strict = True

    fill_control_list = fields.List(
        fields.Dict(
            keys = fields.String(),
            values = fields.String(),
            required = True,
        ),
        required = False,
    )
    """
    FS_49_96_50_77 Config-only delegator implementation to compute default values:
    A `list` of `dict`-s with each list item corresponds to the `search_control_list_`.
    Each `dict` controls setting default values (based on already found `data_envelope`-s):
    *   Each key is a prop name to set default value for.
    *   Each value is an expression for prop value which may refer to already prev `envelope_container`-s.
    """

    func_envelope = fields.Nested(func_envelope_desc.dict_schema)

func_config_desc = TypeDesc(
    dict_schema = FuncConfigSchema(),
    ref_name = FuncConfigSchema.__name__,
    dict_example = {
        fill_control_list_: [
            {
            },
            {
            },
            {
                "severity_level": "{envelope_containers[1].data_envelopes[0]['severity_level']}",
                "exit_code": "{envelope_containers[1].data_envelopes[0]['exit_code']}",
            },
        ],
        func_envelope_: func_envelope_desc.dict_example,
    },
    default_file_path = "",
)



from __future__ import annotations

from typing import Union

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.CompScope import CompScope
from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.misc_helper_server import insert_unique_to_sorted_list
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.InitControl import InitControl
from argrelay.runtime_context.InterpContext import (
    InterpContext,
    function_container_ipos_,
)
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
)
from argrelay.schema_config_interp.InitControlSchema import init_control_desc
from argrelay.schema_config_interp.SearchControlSchema import search_control_desc

func_search_control_ = "func_search_control"
"""
This field is automatically populated by `FuncTreeInterpFactory` inside `interp_tree_node_config_dict`.
"""

func_init_control_ = "func_init_control"
"""
This field is automatically populated by `FuncTreeInterpFactory` inside `interp_tree_node_config_dict`.
"""


class FuncTreeInterp(AbstractInterp):
    """
    Implements FS_26_43_73_72 func tree.

    Finds function `data_envelope` within func tree first,
    then uses its delegator (see `AbstractDelegator`) to find all args-related `data_envelope`-s.

    See FS_55_57_45_04 enum selector.
    """

    def __init__(
        self,
        interp_factory_id,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
        func_ids_to_func_rel_paths: dict[str, list[list[str]]],
        paths_to_jump: dict[tuple[str, ...], tuple[str, ...]],
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = paths_to_jump

        # Allocate first container for function `data_envelope`:
        self.base_container_ipos += 1
        self.interp_ctx.envelope_containers.append(EnvelopeContainer(SearchControl()))

        self.select_next_container()
        self._apply_func_init_control()
        self._apply_func_search_control()

        self.func_ids_to_func_rel_paths: dict[str, list[list[str]]] = func_ids_to_func_rel_paths

    def _apply_func_init_control(self):
        self.interp_ctx.curr_container.assigned_types_to_values[
            ReservedArgType.EnvelopeClass.name
        ] = AssignedValue(
            ReservedEnvelopeClass.ClassFunction.name,
            ArgSource.InitValue,
        )
        func_init_control: InitControl = init_control_desc.dict_schema.load(
            self.interp_tree_node_config_dict[func_init_control_],
        )
        for prop_type, prop_value in func_init_control.init_types_to_values.items():
            self.interp_ctx.curr_container.assigned_types_to_values[prop_type] = AssignedValue(
                prop_value,
                ArgSource.InitValue,
            )
            if prop_type in self.interp_ctx.curr_container.remaining_types_to_values:
                del self.interp_ctx.curr_container.remaining_types_to_values[prop_type]

    def _apply_func_search_control(self):
        # Function `search_control` is populated based on
        # tree path (FS_01_89_09_24 interp tree + FS_26_43_73_72 func tree)
        # and plugin config (rather than data found in `data_envelope`):
        self.interp_ctx.curr_container.search_control = search_control_desc.dict_schema.load(
            self.interp_tree_node_config_dict[func_search_control_]
        )

    def consume_key_args(self) -> None:
        # TODO: FS_20_88_05_60 named args: stub
        pass

    def consume_pos_args(self) -> None:
        """
        Scans through `unconsumed_tokens` and tries to match its value against values of each type.
        """

        consumed_token_ipos_list = []
        for unconsumed_token_ipos in self.interp_ctx.unconsumed_tokens:
            unconsumed_token = self.interp_ctx.parsed_ctx.all_tokens[unconsumed_token_ipos]
            # See if token matches any type by value:

            for arg_type, arg_values in self.interp_ctx.curr_container.remaining_types_to_values.items():
                if unconsumed_token in arg_values:
                    self.interp_ctx.curr_container.assigned_types_to_values[arg_type] = AssignedValue(
                        unconsumed_token,
                        ArgSource.ExplicitPosArg,
                    )
                    consumed_token_ipos_list.append(unconsumed_token_ipos)
                    self.interp_ctx.consumed_tokens.append(unconsumed_token_ipos)
                    # TD_76_09_29_31: overlapped
                    # Assign matching unconsumed arg value to the first type it matches (only once):
                    del self.interp_ctx.curr_container.remaining_types_to_values[arg_type]
                    break

        # perform list modifications out of the prev loop:
        for consumed_token_ipos in consumed_token_ipos_list:
            self.interp_ctx.unconsumed_tokens.remove(consumed_token_ipos)

    def try_iterate(self) -> InterpStep:
        """
        Try to consume more args if possible.

        *   If function was found, start with search for its first envelope class.
        *   If curr envelope class is found, move to the next (until all are found).

        :returns:
        *   `InterpStep.NextInterp`: move to next interpreter: curr interpreter is fully satisfied from the args
        *   `InterpStep.NextEnvelope`: call again curr interpreter: still more things to find in the args
        *   `InterpStep.StopAll`: interpreter sees no point to continue the loop (`InterpContext.interpret_command`)
        """

        # We want single `data_envelope` to be found, not zero, not more than one:
        if self.interp_ctx.curr_container.found_count > 1:
            # Too many `data_envelope`-s - stop:
            return InterpStep.StopAll
        elif self.interp_ctx.curr_container.found_count == 1:

            if self.interp_ctx.curr_container_ipos == self.base_container_ipos:
                # This is a function envelope:
                search_control_list: list[SearchControl] = self.get_search_control_list()
                # Create `EnvelopeContainer`-s for every `data_envelope` to find:
                self.interp_ctx.alloc_searchable_containers(search_control_list)

            if self.interp_ctx.is_last_container():
                # Function does not need any envelopes:
                return InterpStep.NextInterp
            else:
                self.select_next_container()
                self.delegate_init_control()
                # Need more args to consume for the next envelope to find:
                return InterpStep.NextEnvelope

        else:
            # No `data_envelope` = nothing to do:
            return InterpStep.StopAll

    def get_search_control_list(self) -> list[SearchControl]:
        delegator_plugin = self.get_func_delegator()

        search_control_list: list[SearchControl] = delegator_plugin.run_search_control(
            self.get_found_func_data_envelope()
        )
        return search_control_list

    def delegate_init_control(self):
        delegator_plugin = self.get_func_delegator()
        delegator_plugin.run_init_control(
            self.interp_ctx.envelope_containers,
            self.interp_ctx.curr_container_ipos,
        )

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def delegate_fill_control(
        self,
    ) -> bool:
        delegator_plugin = self.get_func_delegator()
        if delegator_plugin:
            return delegator_plugin.run_fill_control(
                self.interp_ctx,
            )
        else:
            return False

    def get_found_func_data_envelope(self) -> Union[dict, None]:
        func_envelope = self.interp_ctx.envelope_containers[(
            self.base_container_ipos + function_container_ipos_
        )]
        if func_envelope.found_count == 1:
            return func_envelope.data_envelopes[0]
        return None

    def get_func_delegator(self):
        func_data_envelope = self.get_found_func_data_envelope()
        if func_data_envelope:
            delegator_plugin_instance_id = func_data_envelope[instance_data_][delegator_plugin_instance_id_]
            delegator_plugin: AbstractDelegator = self.interp_ctx.action_delegators[delegator_plugin_instance_id]
            return delegator_plugin
        else:
            # func envelope hasn't been found yet:
            return None

    def select_next_container(self):
        self.interp_ctx.curr_container_ipos += 1
        self.interp_ctx.curr_container = self.interp_ctx.envelope_containers[
            self.interp_ctx.curr_container_ipos
        ]

    def next_interp(self) -> "AbstractInterp":
        delegator_plugin = self.get_func_delegator()
        interp_factory_id = delegator_plugin.run_interp_control(self)
        if interp_factory_id:
            self.select_next_interp_tree_abs_path()
            return self.interp_ctx.create_next_interp(interp_factory_id)
        else:
            return None

    def select_next_interp_tree_abs_path(self):
        if self.interp_ctx.interp_tree_abs_path in self.paths_to_jump:
            # FS_91_88_07_23 jump tree: replace current `interp_tree_abs_path` with another one based on config:
            self.interp_ctx.interp_tree_abs_path = self.paths_to_jump[self.interp_ctx.interp_tree_abs_path]

    def propose_arg_completion(self) -> None:
        for comp_value in self.propose_auto_comp_list():
            insert_unique_to_sorted_list(self.interp_ctx.comp_suggestions, comp_value)

    def propose_auto_comp_list(self) -> list[str]:

        # TODO: FS_20_88_05_60: POC: Either remove it or implement properly: just testing named args:
        if (
            self.interp_ctx.parsed_ctx.tan_token_l_part.endswith(":")
            or
            self.interp_ctx.parsed_ctx.tan_token_r_part.startswith(":")
        ):
            return [
                type_name + SpecialChar.KeyValueDelimiter.value
                for type_name in self.interp_ctx.curr_container.search_control.types_to_keys_dict
                if not type_name.startswith("_")
            ]

        # TODO: FS_23_62_89_43: the logic for both if-s (`if-A` and `if-B`) is identical at the moment - what do we want to improve?

        # TODO: FS_23_62_89_43: if-A:
        if self.interp_ctx.parsed_ctx.comp_scope is CompScope.ScopeInitial:
            if self.interp_ctx.parsed_ctx.tan_token_l_part == "":
                return self.remaining_from_next_missing_type()
            else:
                return self.remaining_from_next_missing_type()

        # TODO: FS_23_62_89_43: if-B:
        if self.interp_ctx.parsed_ctx.comp_scope is CompScope.ScopeSubsequent:
            if self.interp_ctx.parsed_ctx.tan_token_l_part == "":
                return self.remaining_from_next_missing_type()
            else:
                # TODO: FS_20_88_05_60: Suggest keys (:) of missing types instead - it is `ScopeSubsequent`, user insist and wants something else:
                return self.remaining_from_next_missing_type()

        return []

    def remaining_from_next_missing_type(self) -> list[str]:
        """
        Clarifications:
        *   remaining = because values for the given type are reduced based on narrowed down `data_envelope` set
        *   missing = because this arg type is not specified yet
        *   next = because arg types are tired in specific order
        """
        proposed_values: list[str] = []

        # Return filtered value set from the next missing arg:
        for arg_type in self.interp_ctx.curr_container.search_control.types_to_keys_dict:
            if (
                # TODO: only one condition should be enough: arg_type is either in one or in another, not in both:
                arg_type not in self.interp_ctx.curr_container.assigned_types_to_values
                and
                arg_type in self.interp_ctx.curr_container.remaining_types_to_values
            ):
                proposed_values = [
                    # FS_71_87_33_52: `help_hint`:
                    self.interp_ctx.help_hint_cache.get_value_with_help_hint(arg_type, x)
                    for x in self.interp_ctx.curr_container.remaining_types_to_values[arg_type]
                    if (
                        # NOTE: This checks for `str` meaning primitive (scalar, not `list` or `dict`),
                        #       but this also filters out `int` or others.
                        #       Primitive types should be converted to `str` when loaded.
                        #       See `ConfigOnlyLoader.convert_envelope_fields_to_string` for example.
                        isinstance(x, str)
                        and
                        # FS_32_05_46_00: using `startswith`:
                        # FS_23_62_89_43: filter using L part of tangent token:
                        x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                        # TODO: FS_06_99_43_60: Support list[str] - what if one type can have list of values (and we need to match any as in OR)?
                    )
                ]
                if proposed_values:
                    # Collect only until the first proposed value set from missing args:
                    break

        return proposed_values



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class SearchControl:
    """
    Implements FS_31_70_49_15: search control:
    Object to specify what arg types will be used in search (and how they are mapped to named arg keys).
    """

    collection_name: str = field(default_factory = lambda: None)
    """
    MongoDB collection to search `data_envelope`-s in.
    """

    envelope_class: str = field(default_factory = lambda: None)
    """
    Specifies `ReservedArgType.EnvelopeClass` to search.

    Provides key-value pair (`ReservedArgType.EnvelopeClass`, `envelope_class`) to be fixed in query.
    If this field was not here, it would potentially become an issue because
    user would need to be interrogated on command line to disambiguate `data_envelope`-s with same properties.
    """

    # A specs how to query `data_envelope` for this `EnvelopeContainer` (which arg types to use).
    # Direct list to preserve order:
    keys_to_types_list: list[dict[str, str]] = field(default_factory = lambda: [])

    # `keys_to_types_dict` is derived from `keys_to_types_list`:
    # Direct dict for quick lookup:
    keys_to_types_dict: dict[str, str] = field(init = False, default_factory = lambda: {})

    # `types_to_keys_dict` is derived from `keys_to_types_dict`:
    # Reverse lookup:
    types_to_keys_dict: dict[str, str] = field(init = False, default_factory = lambda: {})

    def __post_init__(self):
        self._init_derived_fields()

    def _init_derived_fields(self):
        self.keys_to_types_dict = self.convert_list_of_ordered_singular_dicts_to_unordered_dict(
            self.keys_to_types_list
        )

        # generate reverse:
        self.types_to_keys_dict = {
            v: k for k, v in
            self.keys_to_types_dict.items()
        }

    @staticmethod
    def convert_list_of_ordered_singular_dicts_to_unordered_dict(dict_list: list[dict]) -> dict:
        """
        Convert ordered list[dict] (with dict having single { key: value }) into unordered dict { keys: values }.
        """
        converted_dict = {}
        for key_to_value_dict in dict_list:
            curr_key = next(iter(key_to_value_dict))
            curr_value = key_to_value_dict[curr_key]
            converted_dict[curr_key] = curr_value

        return converted_dict



from __future__ import annotations

import copy
import os

from git import Repo

from argrelay.custom_integ.GitRepoArgType import GitRepoArgType
from argrelay.custom_integ.GitRepoDelegator import repo_root_abs_path_
from argrelay.custom_integ.GitRepoEntryConfigSchema import (
    repo_rel_path_,
    envelope_properties_,
    is_repo_enabled_,
)
from argrelay.custom_integ.GitRepoEnvelopeClass import GitRepoEnvelopeClass
from argrelay.custom_integ.GitRepoLoaderConfigSchema import (
    git_repo_loader_config_desc,
    load_repo_commits_,
    repo_entries_,
)
from argrelay.custom_integ.git_utils import is_git_repo
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.misc_helper_common import eprint
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import init_envelop_collections
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_id_,
    envelope_payload_,
)


class GitRepoLoader(AbstractLoader):
    """
    Implements FS_67_16_61_97 git_plugin.
    """

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return git_repo_loader_config_desc.dict_from_input_dict(plugin_config_dict)

    def update_static_data(
        self,
        static_data: StaticData,
    ) -> StaticData:

        if not self.plugin_config_dict:
            return static_data

        static_data = self.load_git_objects(static_data)

        return static_data

    def load_git_objects(
        self,
        static_data: StaticData,
    ) -> StaticData:
        """
        Scan `base_path` recursively and load metadata about all Git repos found.
        """

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        class_names = [
            GitRepoEnvelopeClass.ClassGitRepo.name,
            GitRepoEnvelopeClass.ClassGitCommit.name,
        ]

        init_envelop_collections(
            self.server_config,
            class_names,
            # Same index fields for all collections (can be fine-tuned later):
            lambda collection_name, class_name: [enum_item.name for enum_item in GitRepoArgType]
        )

        repo_envelopes = static_data.envelope_collections[
            class_to_collection_map[GitRepoEnvelopeClass.ClassGitRepo.name]
        ].data_envelopes
        commit_envelopes = static_data.envelope_collections[
            class_to_collection_map[GitRepoEnvelopeClass.ClassGitCommit.name]
        ].data_envelopes

        # List of registered Git abs paths:
        repo_root_abs_paths = []
        for repo_base_abs_path in self.plugin_config_dict[repo_entries_]:

            repo_entries = self.plugin_config_dict[repo_entries_][repo_base_abs_path]
            repo_base_abs_path = os.path.expanduser(repo_base_abs_path)

            # Process repo entries collecting `repo_root_rel_path` and `repo_root_abs_path`:
            for repo_entry in repo_entries:

                repo_root_rel_path: str = repo_entry[repo_rel_path_]
                repo_root_abs_path: str = os.path.join(repo_base_abs_path, repo_root_rel_path)

                if not repo_entry[is_repo_enabled_]:
                    eprint(f"INFO: disabled repo: {repo_root_abs_path}")
                    continue

                # Deduplicate Git root paths:
                if repo_root_abs_path not in repo_root_abs_paths:
                    # Register `repo_root_abs_path`:
                    repo_root_abs_paths.append(repo_root_abs_path)
                else:
                    eprint(f"ERROR: duplicate Git repo: {repo_root_abs_path}")
                    raise RuntimeError

                # Query Git root path of curr dir:
                if is_git_repo(
                    repo_root_abs_path,
                ):
                    pass
                else:
                    eprint(f"ERROR: not a Git repo: {repo_root_abs_path}")
                    # Walked into dir outside of Git repo:
                    continue

                repo_root_base_name = os.path.basename(repo_root_abs_path)

                # Produce path component list, for example, if:
                # repo_root_rel_path = "rel/path/to/git/repo.git/"
                # then:
                # path_comp_list = [ "rel", "path", "to", "git", "repo" ]
                path_comp_list = []
                for rel_git_path_comp in repo_root_rel_path.split(os.sep)[:-1]:
                    if rel_git_path_comp not in path_comp_list:
                        path_comp_list.append(rel_git_path_comp)

                ############################################################################################################
                # repos

                repo_envelope: dict = copy.deepcopy(repo_entry[envelope_properties_])

                repo_envelope.update({
                    envelope_id_: repo_root_abs_path,
                    envelope_payload_: {
                        repo_root_abs_path_: repo_root_abs_path,
                    },
                    ReservedArgType.EnvelopeClass.name: GitRepoEnvelopeClass.ClassGitRepo.name,
                    GitRepoArgType.GitRepoRootRelPath.name: repo_root_rel_path,
                    GitRepoArgType.GitRepoRootAbsPath.name: repo_root_abs_path,
                    GitRepoArgType.GitRepoRootBaseName.name: repo_root_base_name,
                    GitRepoArgType.GitRepoPathComp.name: path_comp_list,
                })
                print(repo_envelope)
                repo_envelopes.append(repo_envelope)

                if not self.plugin_config_dict[load_repo_commits_]:
                    continue

                git_repo = Repo(repo_root_abs_path)
                for git_commit in git_repo.iter_commits():
                    ########################################################################################################
                    # commits

                    commit_envelope: dict = copy.deepcopy(repo_entry[envelope_properties_])

                    commit_envelope.update({
                        envelope_id_: f"{repo_root_abs_path}:{git_commit.hexsha}",
                        envelope_payload_: {
                        },
                        ReservedArgType.EnvelopeClass.name: GitRepoEnvelopeClass.ClassGitCommit.name,
                        GitRepoArgType.GitRepoRootRelPath.name: repo_root_rel_path,
                        GitRepoArgType.GitRepoRootAbsPath.name: repo_root_abs_path,
                        GitRepoArgType.GitRepoRootBaseName.name: repo_root_base_name,
                        GitRepoArgType.GitRepoPathComp.name: path_comp_list,
                        GitRepoArgType.GitRepoCommitId.name: git_commit.hexsha,
                        GitRepoArgType.GitRepoCommitAuthorName.name: git_commit.author.name,
                        GitRepoArgType.GitRepoCommitAuthorEmail.name: git_commit.author.email,
                        GitRepoArgType.GitRepoCommitMessage.name: git_commit.message,
                    })
                    commit_envelopes.append(commit_envelope)

        return static_data
